# -*- coding: utf-8 -*-
"""
Copyright (C) 2019 Event-driven Perception for Robotics
Authors: Sim Bamford
Code contributions from:
        Vadim Tikhanoff
This program is free software: you can redistribute it and/or modify it under 
the terms of the GNU General Public License as published by the Free Software 
Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY 
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
PARTICULAR PURPOSE.  See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with 
this program. If not, see <https://www.gnu.org/licenses/>.

Intended as part of bimvee (Batch Import, Manipulation, Visualisation and Export of Events etc)
exportIitYarp takes a data dict resulting from importAe.
Exports data.log and info.log for each channel contained.
Uses the "exportFilePath" parameter for the top-level folder,
creating it if it doesn't exist. This can be relative or absolute.
Defaults to the current directory.
Uses the "yarpOutputPort" parameter for the port defined in info.log.
Default is "/file" - then the following is appended: "/AE:o"
Where more than one channel is contained, the channel name is used in place of "AE".
If the "viewerApp" parameter is True (default = True), then a play.xml 
file is created at the top level directory, which is configured so that when
run from yarpmanager, the data which has been exported can be viewed.
"writeProtected" (default=True) protects any existing data in the target exportFilePath from being overwritten.
"writeMode" (default='w') can be set to append mode ('a') for exporting (large) files in batches.
"""

# %%

import os
import numpy as np
from tqdm import tqdm
import string
import re


# TODO: this could come from header info in some cases
def getDims(dataTypeDict):
    if 'pol' in dataTypeDict:  # dvs
        minX = np.min(dataTypeDict['x'])
        maxX = np.max(dataTypeDict['x'])
        minY = np.min(dataTypeDict['y'])
        maxY = np.max(dataTypeDict['y'])
    elif 'frames' in dataTypeDict:
        minX = 0
        maxX = (dataTypeDict['frames'][0].shape)[0]
        minY = 0
        maxY = (dataTypeDict['frames'][0].shape)[1]
    return minX, maxX, minY, maxY


def writeModule(xmlFile, name, **kwargs):
    xmlFile.write('<module>\n')
    xmlFile.write('    <name> ' + name + ' </name>\n')
    parameters = kwargs.get('parameters', '')
    xmlFile.write('    <parameters> ' + parameters + ' </parameters>\n')
    dependencies = kwargs.get('dependencies', '')
    if dependencies != '':
        xmlFile.write('    <dependencies>\n')
        xmlFile.write('        ' + dependencies + '\n')
        xmlFile.write('    </dependencies>\n')
    xmlFile.write('    <node> localhost </node>\n')
    xmlFile.write('</module>\n')
    xmlFile.write('\n')


def writeConnection(xmlFile, fromPort, toPort, **kwargs):
    xmlFile.write('<connection>\n')
    xmlFile.write('    <from> ' + fromPort + ' </from>\n')
    xmlFile.write('    <to> ' + toPort + ' </to>\n')
    protocol = kwargs.get('protocol', 'udp')
    xmlFile.write('<protocol> ' + protocol + ' </protocol>\n')
    xmlFile.write('</connection>\n')
    xmlFile.write('\n')


def quoting(match):
    matchedString = (match.string[match.span()[0]:match.span()[1]])
    if matchedString == b'\0':
        return b'\\0'
    if matchedString == b'\n':
        return b'\\n'
    if matchedString == b'\r':
        return b'\\r'


def toStringNested(in_string):
    return re.sub(b'([\n\0\r])', quoting, re.sub(b'([\"\\\\])', b'\\\\\g<1>', in_string))


def exportIitYarpViewer(importedDict, **kwargs):
    exportFilePath = kwargs.get('exportFilePath', './')
    with open(os.path.join(exportFilePath, 'play.xml'), 'w') as xmlFile:
        xmlFile.write('<application>\n')
        xmlFile.write('<name>play </name>\n')
        xmlFile.write('\n')
        writeModule(xmlFile=xmlFile, name='yarpdataplayer')
        pathForPlayback = kwargs.get('pathForPlayback', exportFilePath)
        # dependencyStr = ''
        dependencyStr = '<port timeout="100.0" request="load ' + pathForPlayback + '" reply="[ok]">/yarpdataplayer/rpc:i</port>'
        dataTypesToExport = kwargs.get('dataTypes')
        for channelKey in importedDict['data']:
            for dataType in importedDict['data'][channelKey]:
                if dataTypesToExport is not None and dataType not in dataTypesToExport: continue
                channelKeyStripped = str(channelKey).translate(str.maketrans('', '', string.punctuation))
                channelAndDataType = channelKeyStripped + dataType

                if dataType == 'dvs':
                    minX, maxX, minY, maxY = getDims(importedDict['data'][channelKey][dataType])
                    height = maxY + 1
                    width = maxX + 1
                    paramsFramer = '--name /framer' + channelAndDataType + \
                                   ' --height ' + str(height) + \
                                   ' --width ' + str(width) + \
                                   ' --displays "(/' + channelAndDataType + ' (AE ISO))"'
                    writeModule(xmlFile=xmlFile, name='vFramerLite', parameters=paramsFramer,
                                dependencies=dependencyStr)
                    dependencyStr = ''
                    paramsViewer = '--name /viewer/' + channelAndDataType + \
                                   ' --w ' + str((width) * 2) + \
                                   ' --h ' + str((height) * 2) + ' --synch'
                    writeModule(xmlFile=xmlFile, name='yarpview', parameters=paramsViewer)

                    fromPort = '/file/' + channelAndDataType + ':o'
                    toPort = '/framer' + channelAndDataType + '/' + channelAndDataType + '/AE:i'
                    writeConnection(xmlFile=xmlFile, fromPort=fromPort, toPort=toPort)

                    fromPort = '/framer' + channelAndDataType + '/' + channelAndDataType + '/image:o'
                    toPort = '/viewer/' + channelAndDataType
                    writeConnection(xmlFile=xmlFile, fromPort=fromPort, toPort=toPort)

                if dataType == 'frame':
                    minX, maxX, minY, maxY = getDims(importedDict['data'][channelKey][dataType])
                    height = maxY + 1
                    width = maxX + 1
                    paramsViewer = '--name /viewer/' + channelAndDataType + \
                                   ' --w ' + str((width) * 2) + \
                                   ' --h ' + str((height) * 2) + ' --synch' + \
                                   ' --RefreshTime 33'
                    writeModule(xmlFile=xmlFile, name='yarpview', parameters=paramsViewer, dependencies=dependencyStr)
                    dependencyStr = ''

                    fromPort = '/file/' + channelAndDataType + ':o'
                    toPort = '/viewer/' + channelAndDataType
                    writeConnection(xmlFile=xmlFile, fromPort=fromPort, toPort=toPort)

                if dataType == 'imu':
                    paramsFramer = '--name /framer' + channelAndDataType + \
                                   ' --height 200' + \
                                   ' --width 200' + \
                                   ' --displays "(/' + channelAndDataType + ' (AE IMU))"'
                    writeModule(xmlFile=xmlFile, name='vFramerLite', parameters=paramsFramer,
                                dependencies=dependencyStr)
                    dependencyStr = ''
                    paramsViewer = '--name /viewer/' + channelAndDataType + \
                                   ' --w 400 --h 400 --synch'
                    writeModule(xmlFile=xmlFile, name='yarpview', parameters=paramsViewer)

                    fromPort = '/file/' + channelAndDataType + ':o'
                    toPort = '/framer' + channelAndDataType + '/' + channelAndDataType + '/IMUS:i'
                    writeConnection(xmlFile=xmlFile, fromPort=fromPort, toPort=toPort)

                    fromPort = '/framer' + channelAndDataType + '/' + channelAndDataType + '/image:o'
                    toPort = '/viewer/' + channelAndDataType
                    writeConnection(xmlFile=xmlFile, fromPort=fromPort, toPort=toPort)

        xmlFile.write('</application>\n')


def encodeEvents24Bit(ts, x, y, pol, ch=None, **kwargs):
    # timestamps(ts) are 32 bit integers counted with clock period defined by the type of sensor. 
    # Events are encoded as 32 bits with x,y,channel(ch)(c) and polarity(pol)(p) as shown below
    # 0000 00CS TCYY YYYY YYYY XXXX XXXX XXXP
    # t = 0 to indicate events
    # Ignoring channel though
    clock_period_ns = kwargs.get('clock')
    if clock_period_ns is None:
        clock_period_s = 1 / 1e6
    else:
        clock_period_s = clock_period_ns / 1e9

    ts = ts / clock_period_s
    ts = ts.astype(np.uint32)  # Timestamp wrapping occurs here
    ts = np.expand_dims(ts, 1)
    x = x.astype(np.uint32)
    y = y.astype(np.uint32)
    pol[pol<0]=0
    pol = pol.astype(bool)
    pol = (~pol).astype(np.uint32)
    # if ch stream is absent, data has already been split into left and right channels
    # in that case, assign channel value of the whole stream (left or right) from kwargs
    if ch is None:
        ch = kwargs.get('channel', 0)
        ch = np.uint32(ch)
    else:
        ch = ch.astype(np.uint32)
    ae = (ch << 22) + (y << 12) + (x << 1) + pol
    ae = np.expand_dims(ae, 1)
    data = np.concatenate([ts, ae], axis=1)
    return data.flatten()


def exportDvs(dataFile, data, bottleNumber, **kwargs):
    # preconvert numpy arrays into bottle-ready data
    print('Converting event arrays to bottle-ready string format ...')
    ch = None
    if 'ch' in data:
        ch = data['ch']
    eventsAsListOfInts = encodeEvents24Bit(data['ts'],
                                           data['x'],
                                           data['y'],
                                           data['pol'],
                                           ch,
                                           **kwargs)

    print('Breaking into bottles by time')
    # Output dvs data bottle by bottle.
    minTimeStepPerBottle = kwargs.get('minTimeStepPerBottle', 2e-3)
    numEvents = len(data['ts'])
    ptr = 0

    pbar = tqdm(total=numEvents, position=0, leave=True)
    exportAsEv2 = kwargs.get('exportAsEv2', True)
    while ptr < numEvents:
        firstTs = data['ts'][ptr]
        nextPtr = np.searchsorted(data['ts'], firstTs + minTimeStepPerBottle)
        pbar.update(nextPtr - ptr)
        if exportAsEv2:
            eventData = eventsAsListOfInts[ptr * 2: nextPtr * 2]
            if not kwargs.get('exportTimestamps', True):
                eventData = eventData[1::2]

            eventData = toStringNested(eventData.tobytes())
            packetDuration = int(eventsAsListOfInts[::2][min(len(eventsAsListOfInts[::2]) - 1, nextPtr)]) - \
                             int(eventsAsListOfInts[::2][ptr])
            dataFile.write(bytes(f'{bottleNumber} {firstTs:.6f} AE {packetDuration} "', 'utf') + eventData + b'\"\n')
        else:
            eventData = ' '.join(eventsAsListOfInts[ptr * 2: nextPtr * 2].astype(str))
            dataFile.write(f'{bottleNumber} {firstTs:.6f} AE ({eventData})\n')
        ptr = nextPtr
        bottleNumber += 1


def exportFrame(dataFile, data, **kwargs):
    # Here is an example from the python camera - rgb though:
    # 1368 1571751370.660680 [mat] [rgb] (3 230400 8 320 240) {45 43 ... }
    lineStrs = []
    for bottleNum, tsFrameTuple in enumerate(tqdm(list(zip(data['ts'], data['frames'])), position=0, leave=True)):
        ts, frame = tsFrameTuple
        xDim, yDim = frame.shape
        lineStrs.append(str(bottleNum + 1) + ' ' + "{0:.3f}".format(ts) + ' [mat] [mono] (' +
                        ' '.join(['1', str(frame.size), '8', str((frame.shape)[0]), str((frame.shape)[1])]) + ') {' +
                        ' '.join(list(map(str, frame.flatten().tolist()))) + '}\n')

        ''' expand mono to rgb
        frameRgb = np.reshape(frame, (-1, 1))
        frameRgb = (frameRgb.astype('float64') * 99 / 255).astype('uint8')  
        frameRgb = np.concatenate((frameRgb, frameRgb, frameRgb), axis=1).flatten()
        frameRgb = np.random.randint(low=0, high=100, size=129600, dtype=np.uint8)
        lineStrs.append(str(bottleNum + 1) + ' ' + "{0:.6f}".format(ts) + ' [mat] [rgb] (' + 
            ' '.join(['3', str(frame.size * 3), '8', str((frame.shape)[0]), str((frame.shape)[1])]) + ') {' +
            ' '.join(list(map(str, frameRgb.tolist()))) + '}\n')
        '''
    dataFile.write(''.join(lineStrs))


def exportPose6q(dataFile, data, **kwargs):
    # Following Prashanth's lead here:  https://github.com/robotology/whole-body-estimators/blob/b55be34ce0abab46a14b7bb25d39d9ede5d05d0e/devices/baseEstimatorV1/include/baseEstimatorV1.h#L421
    pass


def encodeImu(ts, **kwargs):
    # An IMU reading should go in a bottle with 10 consecutive events.
    # Bottles may contain multiple packets like this
    # An event is a 4-byte ts followed by a 4-byte sample
    # timestamps(ts) are 32 bit integers counted with an 80 ns clock. 
    # Events are encoded as 32 bits with value(v), sensor(s), channel(c), and typet) as shown below
    # rrrr rrrr tcrr ssss vvvv vvvv vvvv vvvv    (r = reserved = 0)
    # Ignoring channel though
    ''' Struct of imu sample is (from LSB to MSB):
    int value:16;
    unsigned int sensor:4; - see below ...
    unsigned int _r1:2; - reserved
    unsigned int channel:1; - left vs right
    unsigned int type:1; - should be 1 to indicate sample (cf event)
    unsigned int _r2:8; - reserved
    'Sensor' is defined as: 
    0: Accel -Y
    1: Accel X
    2: Accel Z
    3: Gyro -Y
    4: Gyro X
    5: Gyro Z
    6: Temperature
    7: Mag -Y
    8: Mag X
    9: Mag Z
    
    For the IMU on STEFI, which is this one:
    ICM-20648, gyro full scale is +/-2000 degrees per second,
    accelerometer full scale is +/-2 g.    
    temp - 333.87 - but does it zero at 0K or at room temperature? (+21deg)
    mag - no idea - this model doesn't have an internal mag
    '''
    numSamples = len(ts)
    accConversionFactor = kwargs.get('accConversionFactor', (32768 / 2) / 9.80665)
    angVConversionFactor = kwargs.get('angVConversionFactor', (32768 / 250) * (180 / np.pi))
    tempConversionFactor = kwargs.get('tempConversionFactor', 333.87)
    tempConversionOffset = kwargs.get('tempConversionOffset', -273.15 - 21)
    magConversionFactor = kwargs.get('magConversionFactor', (32768 / 4900) * 1000000)

    acc = kwargs.get('acc')
    angV = kwargs.get('angV')
    temp = kwargs.get('temp')
    mag = kwargs.get('mag')

    acc = np.zeros((numSamples, 3), dtype=int) if acc is None else (acc * accConversionFactor).astype(int)
    angV = np.zeros((numSamples, 3), dtype=int) if angV is None else (angV * angVConversionFactor).astype(int)
    temp = np.zeros((numSamples, 1), dtype=int) if temp is None else (
            (temp + tempConversionOffset) * tempConversionFactor).astype(int)
    mag = np.zeros((numSamples, 3), dtype=int) if mag is None else (mag * magConversionFactor).astype(int)

    # switch X and Y; negate Y, to match IMU as mounted on Stefi
    acc = acc[:, [1, 0, 2]]
    angV = angV[:, [1, 0, 2]]
    mag = mag[:, [1, 0, 2]]
    acc[:, 0] = - acc[:, 0]
    angV[:, 0] = - angV[:, 0]
    mag[:, 0] = - mag[:, 0]

    ts = ts / 0.00000008  # TODO find way to determine this rescaling value automatically
    ts = ts.astype(np.uint32)  # Timestamp wrapping occurs here
    ts = np.tile(ts, (1, 10))
    ts = ts.flatten(order='C')
    ts = np.expand_dims(ts, 1)
    values = np.concatenate((acc, angV, temp, mag), axis=1)
    values = values.flatten(order='C')
    sensor = np.tile(np.arange(10, dtype=np.uint32), (numSamples, 1))
    sensor = sensor.flatten(order='C')
    channel = kwargs.get('channel', 0)
    eventType = 1
    ae = (eventType << 23) + (channel << 22) + (sensor << 16) + values
    ae = np.expand_dims(ae, 1)
    data = np.concatenate([ts, ae], axis=1)
    data = data.flatten().tolist()
    data = list(map(str, data))
    return data


def exportImu(dataFile, data, bottleNumber, **kwargs):
    # preconvert numpy arrays into bottle-ready data
    print('Converting event arrays to bottle-ready string format ...')
    eventsAsListOfStrings = encodeImu(ts=data['ts'],
                                      acc=data['acc'],
                                      angV=data['angV'])
    print('Breaking into bottles by time')
    # Output dvs data bottle by bottle.
    minTimeStepPerBottle = kwargs.get('minTimeStepPerBottle', 2e-3)
    numEvents = len(data['ts'])
    ptr = 0
    pbar = tqdm(total=numEvents, position=0, leave=True)
    bottleStrs = []
    while ptr < numEvents:
        firstTs = data['ts'][ptr]
        nextPtr = np.searchsorted(data['ts'], firstTs + minTimeStepPerBottle)
        pbar.update(nextPtr - ptr)
        # Why 20 in the following? Because there are 10 samples per imu and 2 ints per sample that need to be included. 
        bottleStrs.append(str(bottleNumber) + ' ' +
                          '{0:.6f}'.format(firstTs) + ' IMUS (' +
                          ' '.join(eventsAsListOfStrings[ptr * 20: nextPtr * 20]) + ')')
        ptr = nextPtr
        bottleNumber += 1
    dataFile.write('\n'.join(bottleStrs))


def encodeSample(data, **kwargs):
    # Samples are encoded like this
    # An event is a 4-byte ts followed by a 4-byte sample
    # timestamps(ts) are 32 bit integers counted with an 80 ns clock. 
    # Events are encoded as 32 bits with value(v), sensor(s), channel(c), and typet) as shown below
    # rrrr rrrr tcrr ssss vvvv vvvv vvvv vvvv    (r = reserved = 0)
    # Ignoring channel though

    # Unpack
    ts = data['ts']
    sensor = data['sensor'].astype(np.uint32)  # change type to allow shifting up bits
    value = data['value'].astype(np.uint16)  # change to unsigned int to avoid wrapping in the sum below
    ts = ts / 0.00000008  # TODO find way to determine this rescaling value automatically
    ts = ts.astype(np.uint32)  # Timestamp wrapping occurs here
    ts = np.expand_dims(ts, 1)
    channel = kwargs.get('channel', 0)
    eventType = 1
    ae = (eventType << 23) + (channel << 22) + (sensor << 16) + value
    ae = np.expand_dims(ae, 1)
    data = np.concatenate([ts, ae], axis=1)
    data = data.flatten().tolist()
    data = list(map(str, data))
    return data


def exportSample(dataFile, data, bottleNumber, **kwargs):
    # preconvert numpy arrays into bottle-ready data
    print('Converting event arrays to bottle-ready string format ...')
    eventsAsListOfStrings = encodeSample(data, **kwargs)
    print('Breaking into bottles by time')
    # Output dvs data bottle by bottle.
    minTimeStepPerBottle = kwargs.get('minTimeStepPerBottle', 2e-3)
    numEvents = len(data['ts'])
    ptr = 0
    pbar = tqdm(total=numEvents, position=0, leave=True)
    bottleStrs = []
    while ptr < numEvents:
        firstTs = data['ts'][ptr]
        nextPtr = np.searchsorted(data['ts'], firstTs + minTimeStepPerBottle)
        pbar.update(nextPtr - ptr)
        # Why 20 in the following? Because there are 10 samples per imu and 2 ints per sample that need to be included. 
        bottleStrs.append(str(bottleNumber) + ' ' +
                          '{0:.6f}'.format(firstTs) + ' IMUS (' +
                          ' '.join(eventsAsListOfStrings[ptr * 2: nextPtr * 2]) + ')')
        ptr = nextPtr
        bottleNumber += 1
    dataFile.write('\n'.join(bottleStrs))


def encodeSkeleton(inData, **kwargs):
    data = np.array([], dtype=np.uint32)
    for i in range(len(inData['ts'])):
        for key in inData:
            if (key != 'ts'):
                data = np.append(data, inData[key][i].astype(np.uint32))

    data = data.flatten().tolist()
    data = list(map(str, data))
    return data


def exportSkeleton(dataFile, data, bottleNumber, **kwargs):
    # preconvert numpy arrays into bottle-ready data
    print('Converting skeleton arrays to bottle-ready string format ...')
    eventsAsListOfStrings = encodeSkeleton(data, **kwargs)
    print('Breaking into bottles by time')
    # Output skeleton data bottle by bottle.
    minTimeStepPerBottle = kwargs.get('minTimeStepPerBottle', 2e-3)
    numEvents = len(data['ts'])
    ptr = 0
    pbar = tqdm(total=numEvents, position=0, leave=True)
    bottleStrs = []
    while ptr < numEvents:
        firstTs = data['ts'][ptr]
        nextPtr = np.searchsorted(data['ts'], firstTs + minTimeStepPerBottle)
        pbar.update(nextPtr - ptr)
        bottleStrs.append(str(bottleNumber) + ' ' +
                          '{0:.6f}'.format(firstTs) + ' SKLT (' +
                          ' '.join(eventsAsListOfStrings[ptr * 13 * 2: nextPtr * 13 * 2]) + ')')
        ptr = nextPtr
        bottleNumber += 1
    dataFile.write('\n'.join(bottleStrs))


def exportIitYarp(importedDict, **kwargs):
    exportFilePath = kwargs.get('exportFilePath', './')
    if isinstance(importedDict, list):
        for idx, elem in enumerate(importedDict):
            kwargs['exportFilePath'] = exportFilePath + '/' + str(idx)
            exportIitYarp(elem, **kwargs)
        return
    print('exportIitYarp called for data imported from ' +
          str(importedDict['info']) +
          ' targeting folder ' + exportFilePath)
    try:
        os.makedirs(exportFilePath)
    except FileExistsError:  # The folder already exists
        pass
    yarpOutputPort = kwargs.get('yarpOutputPort', '/file')
    channelNames = importedDict['data'].keys()
    dataTypesToExport = kwargs.get('dataTypes')
    for channelName in channelNames:
        if channelName in ['1', 'right']:
            kwargs['channel'] = 1
        else:
            kwargs['channel'] = 0
        dataTypes = importedDict['data'][channelName].keys()
        for dataType in dataTypes:
            if dataTypesToExport is None or dataType in dataTypesToExport:  # , 'pose6', 'imu'
                if dataType not in ['dvs', 'frame', 'imu', 'sample', 'skeleton']:
                    print('unknown datatype')
                    continue  # Exclude unknown datatypes
                channelNameAndDataType = channelName + dataType  # This is used as a new channel name
                channelPath = os.path.join(exportFilePath, str(channelNameAndDataType))
                try:
                    os.mkdir(channelPath)
                except FileExistsError:  # The folder already exists
                    pass
                dirList = os.listdir(channelPath)

                writeMode = kwargs.get('writeMode', 'w')
                bottleNumberStart = 0
                if (kwargs.get('exportAsEv2', True)) and 'b' not in writeMode:
                    writeMode += 'b'
                readMode = 'rb' if (kwargs.get('exportAsEv2', True)) else 'r'
                if 'data.log' in dirList:
                    print('data already exists in export directory for channel and datatype' + str(
                        channelNameAndDataType))
                    if kwargs.get('protectedWrite', True):
                        print('export not performed')
                        continue
                    else:
                        if 'w' in writeMode:
                            print('data is being overwritten!')
                        elif 'a' in writeMode:
                            print('data is being appended!')
                            with open(os.path.join(channelPath, 'data.log'), readMode) as f:
                                bottleNumberStart = sum(1 for _ in f)

                # Write the info.log file
                yarpOutputPortForChannel = yarpOutputPort + '/' + channelNameAndDataType + ":o"
                with open(os.path.join(channelPath, 'info.log'), 'w') as infoFile:
                    infoFile.write('Type: Bottle;\n')
                    # TODO: get the real start time from the 'info', if it exists 
                    # - placeholder 0.0 put in the following line 
                    infoFile.write('[0.0] ' + yarpOutputPortForChannel + ' [connected]\n')
                # Write the data.log file
                with open(os.path.join(channelPath, 'data.log'), writeMode) as dataFile:
                    data = importedDict['data'][channelName][dataType]
                    if dataType == 'dvs':
                        exportDvs(dataFile, data, bottleNumberStart, **kwargs)
                    elif dataType == 'frame':
                        # TODO: handle bottle numbering for writeMode = 'a'
                        exportFrame(dataFile, data, **kwargs)
                    elif dataType == 'imu':
                        exportImu(dataFile, data, bottleNumberStart, **kwargs)
                    elif dataType == 'pose6q':
                        # TODO: handle bottle numbering for writeMode = 'a'
                        exportPose6q(dataFile, data, **kwargs)
                    elif dataType == 'sample':
                        exportSample(dataFile, data, bottleNumberStart, **kwargs)
                    elif dataType == 'skeleton':
                        exportSkeleton(dataFile, data, bottleNumberStart, **kwargs)
            else:
                print("datatype: ", dataType, " not handled yet")
    if kwargs.get('viewerApp', True):
        exportIitYarpViewer(importedDict, **kwargs)



# -*- coding: utf-8 -*-
"""
Copyright (C) 2019 Event-driven Perception for Robotics
Authors: Sim Bamford
This program is free software: you can redistribute it and/or modify it under 
the terms of the GNU General Public License as published by the Free Software 
Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY 
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
PARTICULAR PURPOSE.  See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with 
this program. If not, see <https://www.gnu.org/licenses/>.

Intended as part of bimvee (Batch Import, Manipulation, Visualisation and Export of Events etc)
plotImu takes 'inDict' - a dictionary containing imported IMU data 
(or a higher level container, in which attempts to descend and call itself) 
as created by importAe, and plots against time the various dimensions of the 
IMU samples contained. 
"""
import os, sys
import matplotlib.pyplot as plt
import seaborn as sns

sns.set(palette="colorblind")

# Color Palette for Color Blindness
zesty_palette     = ['#F5793A', '#A95AA1', '#85C0F9', '#0F2080']
retro_palette     = ['#601A4A', '#EE442F', '#63ACBE', '#F9F4EC']
corporate_palette = ['#8DB8AD', '#EBE7E0', '#C6D4E1', '#44749D']

#-----------------------------------------------------------------------------------------------------
def plotImu(inDict, **kwargs):
    if isinstance(inDict, list):
        for inDictInst in inDict:
            plotImu(inDictInst, **kwargs)
        return
    if 'info' in inDict:
        fileName = inDict['info'].get('filePathOrName', '')
        print('plotImu was called for file ' + fileName)
        if not inDict['data']:
            print('The import contains no data.')
            return
        for channelName in inDict['data']:
            channelData = inDict['data'][channelName]
            if 'imu' in channelData and len(channelData['imu']['ts']) > 0:
                kwargs['title'] = ' '.join([fileName, str(channelName)])
                plotImu(channelData['imu'], **kwargs)
            else:
                print('Channel ' + channelName + ' skipped because it contains no polarity data')
        return
    if 'temp' in inDict: 
        numSubplots = 4
    else:
        numSubplots = 3
    fig, allAxes = plt.subplots(numSubplots, 1)
    fig.suptitle(kwargs.get('title', ''))
    axesAcc = allAxes[0]
    axesAcc.plot(inDict['ts'], inDict['acc'][:, 0], 'r')
    axesAcc.plot(inDict['ts'], inDict['acc'][:, 1], 'g')
    axesAcc.plot(inDict['ts'], inDict['acc'][:, 2], 'b')
    axesAcc.set_title('Acceleration (m/s)')
    axesAcc.legend(['x', 'y', 'z'])

    axesAngV = allAxes[1]
    axesAngV.plot(inDict['ts'], inDict['angV'][:, 0], 'r')
    axesAngV.plot(inDict['ts'], inDict['angV'][:, 1], 'g')
    axesAngV.plot(inDict['ts'], inDict['angV'][:, 2], 'b')
    axesAngV.set_title('Angular velocity (rad/s)')
    axesAngV.legend(['x', 'y', 'z'])

    axesMag = allAxes[2]
    axesMag.plot(inDict['ts'], inDict['mag'][:, 0], 'r')
    axesMag.plot(inDict['ts'], inDict['mag'][:, 1], 'g')
    axesMag.plot(inDict['ts'], inDict['mag'][:, 2], 'b')
    axesMag.set_title('Mag (uT)')
    axesMag.legend(['x', 'y', 'z'])

    if 'temp' in inDict: 
        axesTemp = allAxes[3]
        axesTemp.plot(inDict['ts'], inDict['temp'])
        axesTemp.set_title('Temp (K)')


#-----------------------------------------------------------------------------------------------------
def plotImuDistribution(imuDict, unitIMU='FPGA', fig_path=None, fig_name=None, fig_subtitle=None):
    """
    Plot the distribution of the IMU data in imuDict. If specified, save the
    generated figure as fig_name.png at the location defined by fig_path.

    Arguments:
        imuDict {dict} -- dictionary of IMU data (as formatted by bimvee)

    Keyword Arguments:
        unitIMU {str} -- either 'FPGA' or 'SI' (default: {'FPGA'})
        fig_path {string} -- save path for the generated figure (default: {None})
        fig_name {string} -- name of the generated figure (default: {None})
        fig_subtitle {string} -- figure sub-title (default: {None})
    """
    fig = plt.figure(figsize=(14.0, 10.0))
    if isinstance(fig_subtitle, str):
        fig.suptitle("IMU Samples Distribution\n" + fig_subtitle, fontsize=20, fontweight='bold')
    else:
        fig.suptitle("IMU Samples Distribution", fontsize=20, fontweight='bold')

    plt.subplot(3,2,1)
    sns.distplot([v for v, s in zip(imuDict['value'], imuDict['sensor']) if s == 0], bins=100, color=zesty_palette[0])
    plt.title("Accelerometer", fontsize=16, fontweight='bold')
    if unitIMU == 'FPGA':
        plt.xlabel('accX [fpga]', fontsize=10, fontweight='bold')
    elif unitIMU == 'SI':
        plt.xlabel('accX [m/s]', fontsize=10, fontweight='bold')

    plt.subplot(3,2,2)
    sns.distplot([v for v, s in zip(imuDict['value'], imuDict['sensor']) if s == 3], bins=100, color=zesty_palette[0])
    plt.title("Gyroscope", fontsize=16, fontweight='bold')
    if unitIMU == 'FPGA':
        plt.xlabel('gyroX [fpga]', fontsize=10, fontweight='bold')
    elif unitIMU == 'SI':
        plt.xlabel('gyroX [rad/s]', fontsize=10, fontweight='bold')

    plt.subplot(3,2,3)
    sns.distplot([v for v, s in zip(imuDict['value'], imuDict['sensor']) if s == 1], bins=100, color=zesty_palette[1])
    if unitIMU == 'FPGA':
        plt.xlabel('accY [fpga]', fontsize=10, fontweight='bold')
    elif unitIMU == 'SI':
        plt.xlabel('accY [m/s]', fontsize=10, fontweight='bold')

    plt.subplot(3,2,4)
    sns.distplot([v for v, s in zip(imuDict['value'], imuDict['sensor']) if s == 4], bins=100, color=zesty_palette[1])
    if unitIMU == 'FPGA':
        plt.xlabel('gyroY [fpga]', fontsize=10, fontweight='bold')
    elif unitIMU == 'SI':
        plt.xlabel('gyroY [rad/s]', fontsize=10, fontweight='bold')

    plt.subplot(3,2,5)
    sns.distplot([v for v, s in zip(imuDict['value'], imuDict['sensor']) if s == 2], bins=100, color=zesty_palette[3])
    if unitIMU == 'FPGA':
        plt.xlabel('accZ [fpga]', fontsize=10, fontweight='bold')
    elif unitIMU == 'SI':
        plt.xlabel('accZ [m/s]', fontsize=10, fontweight='bold')

    plt.subplot(3,2,6)
    sns.distplot([v for v, s in zip(imuDict['value'], imuDict['sensor']) if s == 5], bins=100, color=zesty_palette[3])
    if unitIMU == 'FPGA':
        plt.xlabel('gyroZ [fpga]', fontsize=10, fontweight='bold')
    elif unitIMU == 'SI':
        plt.xlabel('gyroZ [rad/s]', fontsize=10, fontweight='bold')

    fig.tight_layout()
    if isinstance(fig_subtitle, str):
        fig.subplots_adjust(top=0.85)
    else:
        fig.subplots_adjust(top=0.9)

    if isinstance(fig_path, str) and isinstance(fig_name, str): 
        plt.savefig(os.path.join(fig_path, fig_name + '.png'), dpi=300, bbox_inches='tight')
        print("Saving " + fig_name + ".png")
        plt.close()
    else:
        plt.show()



# -*- coding: utf-8 -*-
"""
Copyright (C) 2020 Event-driven Perception for Robotics
Authors: Sim Bamford
following https://github.com/jpdominguez/pyNAVIS/blob/master/src/pyNAVIS/loaders.py
    by Juan Pedro Dominguez
This program is free software: you can redistribute it and/or modify it under 
the terms of the GNU General Public License as published by the Free Software 
Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY 
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
PARTICULAR PURPOSE.  See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with 
this program. If not, see <https://www.gnu.org/licenses/>.

Intended as part of bimvee (Batch Import, Manipulation, Visualisation and Export of Events etc)
importUdsAedat is a function for importing timestamped address-event data, given
a path or a file containing an aedat file in the format used by PyNavis.
The output is a file-level container in bimvee format.
In general this is a dict containing:
    - info
    - data
    info: will contain minimally:
        - filePathAndName
        - fileFormat
    data: this is a list of dicts, one for each sensor or "channel" which has 
    been imported. Bear in mind that sub-functions may optionally split or join 
    channels. Within each dict, there is a field for each type of data contained.
    Each of these fields is a dict containing a numpy array for timestamps 'ts'
    (one row per event), and a np array for each type of data, with a row for 
    each of those timestamps.
    In particular there will be at least an 'ear' field, containing at least:
        ts            (timestamp)
        freq          (frequency channel, incresing addr -> decreasing frequency)
    and possibly:
        pol           (polarity 0 -> positive; 1 -> negative)
        ch            (channel: 0 -> left; 1 -> right)
        xso           (Olive model: 0 -> MSO; 1 -> LSO)
        auditoryModel (auditory model: 0 -> cochlea; 1 -> superior olivary complex)
        itdNeuron     (Addresses of Interaural-time-difference neurons)
        
There are basic formats corresponding to Ini Aedat v1 and 2.
Files don't necessarily contain headers to disambiguate these, therefore the
user must specify in kwargs any deviations from the following assumptions:
    (loosely following https://github.com/jpdominguez/pyNAVIS/blob/master/src/pyNAVIS/main_settings.py)
    codec: 'Addr2Bytes' i.e. [address(2-bytes) timestamp(4-bytes)] default
              'Addr4Bytes' i.e. [address(4-bytes) timestamp(4-bytes)]
    stereo = True
    splitByChannel = False (i.e. left vs right into separate channel dicts)
    numChannelBits: 5 (number of bits used for frequency channel addresses per sensor)
    tsTick: 1e-6 (i.e. an integer time increment corresponds to 1 microsecond)
    polarised: True (i.e. decode polarity of spikes)
    zeroTime: True (whether to offset timestamps so that they start with zero)
    
    The address is interpretted as: 0...0cf...fp    where:
        p is the polarity bit and may or may not be present
        f are the frequency bits - there are ceil(log_2(numChannels))of these
        c is the channel bit - left vs right, and may or may not be present.
    
"""

#%%

import os
import numpy as np

# local imports
from .timestamps import zeroTimestampsForADataType
from .split import splitByLabel

def importUdsAedat(**kwargs):
    filePathOrName = kwargs.get('filePathOrName')
    if not os.path.exists(filePathOrName):
        raise FileNotFoundError("File or folder not found.")
    # Move forward assuming that it's a single file
    # TODO: implement hierarchical descent through folders
    codec = kwargs.get('codec', 'addr4Bytes')
    addrSize = 4 if codec == 'addr4Bytes' else 2
    importFromByte = kwargs.get('importFromByte', 0)
    importMaxBytes = kwargs.get('importMaxBytes')
    headers = ''
    with open(filePathOrName, 'rb') as file:
        file.seek(importFromByte)

        ## Check header ##
        lt = file.readline()
        while lt and lt[0] == ord("#"):
            headers = headers + lt
            importFromByte += len(lt)
            lt = file.readline()
        file.seek(0, 2)
        eof = file.tell()
        # TODO: headers to info

        # Establish how far ahead to read
        if importMaxBytes is None:
            importMaxBytes = eof - importFromByte
        else:
            importMaxBytes -= np.mod(importMaxBytes, addrSize + 4) # Import a multiple of the word size
            # Check if importMaxBytes is "feasible" (not beyond the file size)?
            if eof > importFromByte + importMaxBytes:
                importMaxBytes = eof - importFromByte
                
        # Now Read from the start point
        file.seek(importFromByte)
        events = file.read(importMaxBytes)
        
        # Pass out where we got to
        if file.read(1) is None:
            kwargs['importedToByte'] = 'EOF'
        else:
            kwargs['importedToByte'] = file.tell() - 2
            # - 2 compensates for the small read-ahead just performed to test EOF

        # Convert to uint16 - it's the highest common denominator for the address format possibilities. 
#           dt = np.dtype(int)
#>>> dt = dt.newbyteorder('>')
#>>> np.frombuffer(buf, dtype=dt) 
 
        events = np.frombuffer(events, np.uint8)
        # The following is one way to recover big-endian order while still using np.frombuffer
        # TODO: feels like there should be a better way
        events = events.reshape((-1, addrSize + 4)).astype(np.uint32)
        tsBytes = events[:, -4:]
        ts = tsBytes[:, -1]
        for byteIdx in range(3):
            ts = ts + tsBytes[:, byteIdx] * 2 ** (8 * (4 - byteIdx))
        #ts = events[:, -2].astype(np.uint32) * 2**16 + events[:, -1]
        earDict = {'ts': ts}
        if addrSize == 4:
            addr = (events[:, 0] * 2**24 +
                    events[:, 1] * 2**16 +
                    events[:, 2] * 2**8 +
                    events[:, 3])
        else:
            addr = (events[:, 0] * 2**8 +
                    events[:, 1])
        if kwargs.get('polarised', ('polarized', True)):
            earDict['pol'] = bool_(addr & 0x01)
            addr = addr >> 1
        numChannelBits = kwargs.get('numChannelBits', 5)
        earDict['freq'] = np.uint16(addr & (2**numChannelBits -1))
        addr >>= numChannelBits
        if kwargs.get('stereo', True):
            earDict['ch'] = np.uint8(addr & 0x01)
#                'itdNeuronIds': itdNeuronIds,
#                'auditoryModel': auditoryModel,
#                'xsoType': xsoType,
#        xsoType = np.uint8(addr & 0x01)
#        addr >>= 1
#        auditoryModel = np.uint8(addr & 0x01)
#        addr >>= 3
#        itdNeuronIds = np.uint8(addr & 0x7F)
#        addr >>= 10
    if kwargs.get('zeroTimestamps', True):
        zeroTimestampsForADataType(earDict) # TODO: This should return a new dict
    tsTick = kwargs.get('tsTick', 1e-6)
    earDict['ts'] = earDict['ts'].astype(float) * tsTick
    if kwargs.get('stereo', True) and kwargs.get('splitByChannel', True):
        channels = splitByLabel(earDict, 'ch')
        if 0 in channels.keys():
            channels['left'] = {'ear': channels.pop(0)}
        if 1 in channels.keys():
            channels['right'] = {'ear': channels.pop(1)}
    else:
        channels = {'general': {'ear': earDict}}

    if headers == '':
        kwargs['headers'] = headers
    outDict = {
        'info': kwargs,
        'data': channels
        }
    outDict['info']['fileFormat'] = 'iityarp'
    return outDict



# -*- coding: utf-8 -*-
"""
Copyright (C) 2020 Event-driven Perception for Robotics
Authors: Simeon Bamford
This program is free software: you can redistribute it and/or modify it under 
the terms of the GNU General Public License as published by the Free Software 
Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY 
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
PARTICULAR PURPOSE.  See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with 
this program. If not, see <https://www.gnu.org/licenses/>.

Intended as part of bimvee (Batch Import, Manipulation, Visualisation and Export of Events etc)
A function which exports a pose6q dataType dict as a csv for use by Rpg's eSim
simulator (https://github.com/uzh-rpg/rpg_esim)

Note:
    a) conversion of ts to ns
    b) switch of quaternion to qx, qy, qz, qw format
"""


def exportPoseRpgEsimCsv(poseDict, filePathAndName='poses.csv'):
    ts = poseDict['ts']
    point = poseDict['point']
    rotation = poseDict['rotation']
    with open(filePathAndName, 'w') as file:
        file.write('# timestamp, x, y, z, qx, qy, qz, qw\n')
        for idx in range(ts.shape[0]):
            file.write('%f, %f, %f, %f, %f, %f, %f, %f\n' % (
                    ts[idx] * 1000000000,                                                       
                    point[idx, 0],                                                       
                    point[idx, 1],                                                       
                    point[idx, 2],                                                       
                    rotation[idx, 1],                                                       
                    rotation[idx, 2],                                                       
                    rotation[idx, 3],                                                       
                    rotation[idx, 0],                                                       
                    ))



# -*- coding: utf-8 -*-

"""
Copyright (C) 2019 Event-driven Perception for Robotics
Authors: Sim Bamford
This program is free software: you can redistribute it and/or modify it under 
the terms of the GNU General Public License as published by the Free Software 
Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY 
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
PARTICULAR PURPOSE.  See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with 
this program. If not, see <https://www.gnu.org/licenses/>.

Intended as part of importRosbag.

The importTopic function receives a list of messages and returns
a dict with one field for each data field in the message, where the field
will contain an appropriate iterable to contain the interpretted contents of each message.
In some cases, static info is repeated in each message; in which case a field may not contain an iterable. 

This function imports the ros message type defined at:

http://docs.ros.org/api/sensor_msgs/html/msg/CameraInfo.html
We assume that there will only be one camera_info msg per channel,
so the resulting dict is populated by the following fields:
    std_msgs/Header header
    uint32 height
    uint32 width
    string distortion_model
    <following as numpy matrices of the appropriate dimensions>
    float64[] D (distortion params)
    float64[9] K (Intrinsic camera matrix)
    float64[9] R (rectification matrix - only for stereo setup)
    float64[12] P (projection matrix)
    <ignoring the following for now:>
    uint32 binning_x
    uint32 binning_y
    sensor_msgs/RegionOfInterest roi

"""

#%%

from .common import unpackRosString, unpackRosUint32, unpackRosFloat64Array

def importTopic(msgs, **kwargs):

    outDict = {}
    data = msgs[0]['data'] # There is one calibration msg per frame. 
    # Just use the first one
    #seq = unpack('=L', data[0:4])[0]
    #timeS, timeNs = unpack('=LL', data[4:12])
    frame_id, ptr = unpackRosString(data, 12)
    outDict['height'], ptr = unpackRosUint32(data, ptr)
    outDict['width'], ptr = unpackRosUint32(data, ptr)
    outDict['distortionModel'], ptr = unpackRosString(data, ptr)
    numElementsInD, ptr = unpackRosUint32(data, ptr)
    outDict['D'], ptr = unpackRosFloat64Array(data, numElementsInD, ptr)
    outDict['K'], ptr = unpackRosFloat64Array(data, 9, ptr)
    outDict['K'] = outDict['K'].reshape(3, 3)
    outDict['R'], ptr = unpackRosFloat64Array(data, 9, ptr)
    outDict['R'] = outDict['R'].reshape(3, 3)
    outDict['P'], ptr = unpackRosFloat64Array(data, 12, ptr)
    outDict['P'] = outDict['P'].reshape(3, 4)
    # Ignore binning and ROI
    return outDict



# -*- coding: utf-8 -*-

"""
Copyright (C) 2020 Event-driven Perception for Robotics
Authors: Ander Arriandiaga Laresgoiti
        Sim Bamford

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with
this program. If not, see <https://www.gnu.org/licenses/>.

Intended as part of bimvee (Batch Import, Manipulation, Visualisation and Export of Events etc)
importProph opens a file in either the .raw or .dat formats of Prophesee. 

https://docs.prophesee.ai/data_formats/file_formats/dat.html
https://docs.prophesee.ai/data_formats/data_encoding_formats/evt2.html

Returns a dict in this format:
{'info': {},
 'data': {
         ch0: {
               dvs: {
                     'ts': np.array of float in seconds
                     'x': np.array of np.uint16 in pixels
                     'y': np.array of np.uint16 in pixels
                     'pol': np.array of bool -- 1 = ON event
                    }}}}
"""

import numpy as np
from struct import unpack

# Local imports
from .timestamps import zeroTimestampsForADataType
from bimvee.importBoundingBoxes import importBoundingBoxes
import os

def importDatHeaders(dat):

    # Go to the beginning of the file
    dat.seek(0)

    info = {}

    # Assume the format version is 1 unless a header of the version number is
    # found
    info['fileFormat'] = 1

    # Read the first character
    is_comment = '%' in str(dat.read(1))
    while is_comment:

        # Read the rest of the line
        line = dat.readline().decode('utf-8', "ignore")
        # File format
        if line[22: 24] == 'CD':
            info['fileFormat'] = line[22: 24]

        # Version
        if line[: 9] == ' Version ':
            info['version'] = line[9:-1]
        # Pick out date and time of recording
        #  % Date YYY-MM-DDHH:MM:SS
        if line[: 6] == ' Date ':
            info['dateTime'] = line[6:-1]

        # Horizontal size of image sensor array
        if line[: 7] == ' Width ':
            info['Width'] = line[7:-1]

        # Horizontal size of image sensor array
        if line[: 8] == ' Height ':
            info['Height'] = line[8:-1]

        # Read ahead the first character of the next line to complete the
        # while loop
        is_comment = '%' in str(dat.read(1))

    # We have read ahead one byte looking for '#', and not found it.
    # Now wind back one to be in the right place to start reading
    dat.seek(-1, 1)

    return info


def importDat(dat):

    # Binary event type & size
    # This information is provided as two bytes.
    # First one indicates event type: for now, only CD type is supported (0x00).
    # The second provide the Event size: always 8
    try:
        event_type = unpack('b', dat.read(1))[0]
        assert event_type == 12
        
        event_size = int(unpack('b', dat.read(1))[0])
        assert event_size == 8
    except AssertionError:
        print('This file format not supported by this decoder.')
        raise
    
    events = dat.read()
    events = np.frombuffer(events, np.uint32().newbyteorder('<'))
    events = events.reshape(-1, 2)
    ts = events[:,0].astype(float) / 1000000
    events = events[:,1]
    x = (events & 0x00003FFF).astype(int)
    y = ((events & 0x0FFFC000) >> 14).astype(int)
    pol = (events & 0xF0000000) >> 28
    
    try:
        assert np.max(pol) < 2 # why are there 4 bits for polarity? Is this for TD events?
    except AssertionError:
        print('This file format not supported by this decoder.')
        raise
    pol = pol.astype(bool)
    
    dvsDict = {'ts': ts,
               'x': x,
               'y': y,
               'pol': pol,
               }
    
    return dvsDict


def importRawHeaders(raw):

    # Go to the beginning of the file
    raw.seek(0)
    info = {}
    # Read the first character
    is_comment = '%' in str(raw.read(1))
    while is_comment:
        # Read the rest of the line
        line = raw.readline().decode('utf-8', "ignore")
        # Pick out date and time of recording
        #  % Date YYY-MM-DDHH:MM:SS
        if line[: 6] == ' Date ':
            info['dateTime'] = line[6:-1]
        # Integrator name
        if line[: 17] == ' integrator_name ':
            info['IntegratorName'] = line[17:-1]
        # Plugin name
        if line[: 13] == ' plugin_name ':
            info['PluginName'] = line[13:-1]
        # Serial Number
        if line[: 15] == ' serial_number ':
            info['SerialNumber'] = line[15:-1]
        # EVT
        if line[: 5] == ' evt ':
            info['evt'] = line[5:-1]
        # Firmware Version
        if line[: 18] == ' firmware_version ':
            info['FirmwareVersion'] = line[18:-1]
        # System ID
        if line[: 11] == ' system_ID ':
            info['SystemID'] = line[11:-1]
        # Subsystem ID
        if line[: 14] == ' subsystem_ID ':
            info['SubsystemID'] = line[14:-1]
        # Horizontal size of image sensor array
        if line[: 7] == ' Width ':
            info['Width'] = line[7:-1]
        # Horizontal size of image sensor array
        if line[: 8] == ' Height ':
            info['Height'] = line[8:-1]
        # Read ahead the first character of the next line to complete the
        # while loop
        is_comment = '%' in str(raw.read(1))
    # We have read ahead one byte looking for '#', and not found it.
    # Now wind back one to be in the right place to start reading
    raw.seek(-1, 1)
    
    return info


def importRaw(raw):
    """ Read .raw binary files which are direct dumps of the ATIS3 USB stream.
    (not kAER _td.dat files)
    Event format is of the evt 2.0, the small footprint ATIS used in ECOMODE.
    
    Beware: this function only returns CD_OFF, CD_ON and LEFT_TD_HIGH events.
    NO EXT_TRIGGER. NO OTHERS.

    Each event is of size 32 bits (or 4 bytes).
    The enclosed information depends on the event type. See doc. """

    # Types of events outputed by the ATIS3
    CD_OFF = 0
    CD_ON = 1
    EVT_TIME_HIGH = 8
    EXT_TRIGGER = 12
    OTHERS = 14
    CONTINUED = 15

    data = raw.read()
    events = np.frombuffer(data, np.uint32().newbyteorder('<'))
    ev_type = (events & 0xF0000000) >> 28
    ts_MSB = (events & 0x0FFFFFFF) << 6  # MSB of the timestamp, shifted by 6 bits to make room for the LSB of the timestamp
    # Spread the TS_MSB events through to the following events
    eventIsTsHigh = ev_type == EVT_TIME_HIGH
    tsHighIds = np.where(eventIsTsHigh)[0]
    tsHighIds = np.append(tsHighIds, len(events))
    for startIdx, endIdx in zip(tsHighIds[:-1], tsHighIds[1:]):
        ts_MSB[startIdx : endIdx] = ts_MSB[startIdx]
    if tsHighIds[0] > 0:
        # wind back the clock by one ts_MSB for the initial events
        # it's not really defined but it's a safe assumption
        ts_MSB[0 : tsHighIds[0]] = ts_MSB[tsHighIds[0]] - 64
    # Eliminate ts_MSB events
    events = events[~eventIsTsHigh]
    ts_MSB = ts_MSB[~eventIsTsHigh]
    ev_type = ev_type[~eventIsTsHigh]

    eventIsOtherType = ev_type > 1
    if np.any(eventIsOtherType):
        print('Other event types are present in this file, which are not being decoded')
    events = events[~eventIsOtherType]
    ts_MSB = ts_MSB[~eventIsOtherType]
    pol = (ev_type[~eventIsOtherType]).astype(bool)
    # unpack the rest of the data
    y = (events & 0x000007FF).astype(int)           # y is bits 10..0
    x = ((events & 0x003FF800) >> 11).astype(int)    # x is bits 21..11
    ts_LSB = (events & 0x0FC00000) >> 22  # ts is bits 27..22
    ts = (ts_MSB + ts_LSB) / 1000000

    dvsDict = {'ts': ts,
        'x': x,
        'y': y,
        'pol': pol,
        }

    return dvsDict



def importProph(**kwargs):
    filePathOrName = kwargs['filePathOrName']
    if filePathOrName[-4:] == '.dat':

        print('Attempting to import ' + filePathOrName + ' as ATIS3 dat file')
        with open(filePathOrName, 'rb') as file:
            file_info = importDatHeaders(file)
            dvsDict = importDat(file)
    elif filePathOrName[-4:] == '.raw':
        print('Attempting to import ' + filePathOrName + ' as ATIS3 USB dumb raw file')
        with open(filePathOrName, 'rb') as file:
            file_info = importRawHeaders(file)
            dvsDict = importRaw(file)

    if kwargs.get('zeroTime', kwargs.get('zeroTimestamps', True)):
        zeroTimestampsForADataType(dvsDict)
    outDict = {
        'info': {'filePathOrName': filePathOrName,
                 'fileFormat': 'Atis3DAT'},
        'data': {
            'ch0': {
                'dvs': dvsDict
            }
        }
    }
    outDict['info'].update(file_info)
    print('Done.')
    gt_file = os.path.join(os.path.dirname(filePathOrName), 'ground_truth.csv')
    if os.path.exists(gt_file):
        outDict['data']['ch0']['boundingBoxes'] = importBoundingBoxes(filePathOrName=gt_file)
    return outDict

# Legacy name for importProph
def importAtis3(**kwargs):
    return importProph(**kwargs)



# -*- coding: utf-8 -*-
"""
Copyright (C) 2019 Event-driven Perception for Robotics
Authors: Sim Bamford
        Massimiliano Iacono

This program is free software: you can redistribute it and/or modify it under 
the terms of the GNU General Public License as published by the Free Software 
Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY 
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
PARTICULAR PURPOSE.  See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with 
this program. If not, see <https://www.gnu.org/licenses/>.

Intended as part of bimvee (Batch Import, Manipulation, Visualisation and Export of Events etc)
Functionality for serving images which represent the data at a certain time
(or given a certain time window). 
The intended use case is to support video-like playback of data
There is a generic Visualiser class - it contains a dataType dict. 
This is subclassed for each supported dataType (e.g. dvs, frame, pose etc)
Each subclass should implement basic methods: get_dims, get_frame, get_colorfmt
Implementing set_data allows the visualiser to do some preparation for visualisation
when it receives new data.

VisualiserPose6q is for visualising pose6q class.
An orientation frame (perpendicular unit vectors of rgb) are generated for each
bodyId (It can handle multiple bodyIds in the same dataDict)
The world (everywhere the trajectories reach in the data) is compressed into
a rectangle which is then viewed from the outside as if the world reference
frame were aligned with a standard camera reference frame,
so z increases away from the viewer, x increases to the right,
and y in a downwards direction.
The user has a choice to apply perspective (scale decreases away from the 
viewer) or not.
There is also a choice of interpolation - if chosen, the pose drawn for a given
time is linearly interpolated between the two nearest poses.
"""

import numpy as np
import math
import cv2

# Local imports
from ..geometry import quat2RotM, slerp
from ..split import splitByLabel
from .visualiserBase import Visualiser


# A function intended to find the nearest timestamp
# adapted from https://stackoverflow.com/questions/2566412/find-nearest-value-in-numpy-array
def findNearest(array, value):
    idx = np.searchsorted(array, value)  # side="left" param is the default
    if idx > 0 and ( \
                    idx == len(array) or \
                    math.fabs(value - array[idx - 1]) < math.fabs(value - array[idx])):
        return idx - 1
    else:
        return idx


# %% Two helper functions for pose visualiser

# adapted from https://stackoverflow.com/questions/50387606/python-draw-line-between-two-coordinates-in-a-matrix
def draw_line(mat, x0, y0, x1, y1):
    if (x0, y0) == (x1, y1):
        if x0 >= 0 and x0 < mat.shape[1] and y0 >= 0 and y0 < mat.shape[0]:
            mat[x0, y0] = 255
        return
    # Swap axes if Y slope is smaller than X slope
    transpose = abs(x1 - x0) < abs(y1 - y0)
    if transpose:
        mat = mat.T
        x0, y0, x1, y1 = y0, x0, y1, x1
    # Swap line direction to go left-to-right if necessary
    if x0 > x1:
        x0, y0, x1, y1 = x1, y1, x0, y0
    # Compute intermediate coordinates using line equation
    x = np.arange(x0, x1 + 1)
    y = np.round(((y1 - y0) / (x1 - x0)) * (x - x0) + y0).astype(x.dtype)
    # Write intermediate coordinates
    toKeep = np.logical_and(x >= 0,
                            np.logical_and(x < mat.shape[1],
                                           np.logical_and(y >= 0, y < mat.shape[0])))
    mat[y[toKeep], x[toKeep]] = 255


def rotateUnitVectors(rotM, unitLength=1.0):
    xVec = np.expand_dims(np.array([unitLength, 0, 0, 1]), axis=1)
    yVec = np.expand_dims(np.array([0, unitLength, 0, 1]), axis=1)
    zVec = np.expand_dims(np.array([0, 0, unitLength, 1]), axis=1)
    allVecs = np.concatenate((xVec, yVec, zVec), axis=1)
    return rotM.dot(allVecs)


class VisualiserPose6q(Visualiser):
    renderX = 200  # TODO Hardcoded
    renderY = 200
    labels = None
    data_type = 'pose6q'

    def __init__(self, data):
        self.set_data(data)
        self.smallestRenderDim = min(self.renderX, self.renderY)

    '''
    Offset and scale the pose translations so that they all fit into the volume:
        x [-0.5:0.5]
        y[-0.5:0.5]
        z[1:2]
    '''

    def set_data(self, data):
        # scale and offset point data so that it remains proportional 
        # but stays in the range 0-1 for all dimensions
        pointX = data['point'][:, 0]
        pointY = data['point'][:, 1]
        pointZ = data['point'][:, 2]
        minX = np.min(pointX)
        maxX = np.max(pointX)
        minY = np.min(pointY)
        maxY = np.max(pointY)
        minZ = np.min(pointZ)
        maxZ = np.max(pointZ)
        centreX = (minX + maxX) / 2
        centreY = (minY + maxY) / 2
        centreZ = (minZ + maxZ) / 2
        largestDim = max(maxX - minX, maxY - minY, maxZ - minZ)
        if largestDim == 0:
            largestDim = 1

        pointScaled = np.empty_like(data['point'])
        pointScaled[:, 0] = pointX - centreX
        pointScaled[:, 1] = pointY - centreY
        pointScaled[:, 2] = pointZ - centreZ
        pointScaled = pointScaled / largestDim + 0.5
        internalData = {'ts': data['ts'],
                        'point': pointScaled,
                        'rotation': data['rotation']}
        # Having scaled data for all poses, split out the poses for different bodies, if applicable
        if 'bodyId' in data:
            # split pose data by label, for ease of reference during rendering
            internalData['bodyId'] = data['bodyId']
            self.__data = splitByLabel(internalData, 'bodyId', outList=False)
            self.labels = np.unique(data['bodyId'])
        else:
            self.__data = {'': internalData}

    def project3dTo2d(self, x=0, y=0, z=0, **kwargs):
        smallestRenderDim = kwargs.get('smallestRenderDim', 1)
        windowFill = kwargs.get('windowFill', 0.9)
        if kwargs.get('perspective', True):
            # Move z out by 1, so that the data is between 1 and 2 distant in z
            # x and y are in range 0-1, so they get shifted to be centred around 0 during this operation
            x = (x - 0.5) / (z + 1) + 0.5
            y = (y - 0.5) / (z + 1) + 0.5
        x = (x * windowFill + (1 - windowFill) / 2) * smallestRenderDim
        y = (y * windowFill + (1 - windowFill) / 2) * smallestRenderDim
        x = x.astype(int)
        y = y.astype(int)
        return x, y

    def project_pose(self, point, rotation, image, **kwargs):
        if point is None:
            return image, None
        # Unpack
        pointX = point[0]
        pointY = point[1]
        pointZ = point[2]
        # Project the location
        kwargs['smallestRenderDim'] = self.smallestRenderDim
        projX, projY = self.project3dTo2d(x=pointX, y=pointY, z=pointZ, **kwargs)
        if rotation is None:
            # Just put a white dot where the point should be
            image[projY, projX, :] = 255
        else:
            rotMats = quat2RotM(rotation)
            rotatedUnitVectors = rotateUnitVectors(rotMats, unitLength=0.25)
            # aVectorCoordB means the Bth coordinate of unit vector originally directed towards a
            xVectorCoordX = pointX + rotatedUnitVectors[0, 0]
            xVectorCoordY = pointY + rotatedUnitVectors[1, 0]
            xVectorCoordZ = pointZ + rotatedUnitVectors[2, 0]
            yVectorCoordX = pointX + rotatedUnitVectors[0, 1]
            yVectorCoordY = pointY + rotatedUnitVectors[1, 1]
            yVectorCoordZ = pointZ + rotatedUnitVectors[2, 1]
            zVectorCoordX = pointX + rotatedUnitVectors[0, 2]
            zVectorCoordY = pointY + rotatedUnitVectors[1, 2]
            zVectorCoordZ = pointZ + rotatedUnitVectors[2, 2]
            xVectorProjX, xVectorProjY = self.project3dTo2d(x=xVectorCoordX, y=xVectorCoordY, z=xVectorCoordZ, **kwargs)
            yVectorProjX, yVectorProjY = self.project3dTo2d(x=yVectorCoordX, y=yVectorCoordY, z=yVectorCoordZ, **kwargs)
            zVectorProjX, zVectorProjY = self.project3dTo2d(x=zVectorCoordX, y=zVectorCoordY, z=zVectorCoordZ, **kwargs)
            draw_line(image[:, :, 0], projX, projY, xVectorProjX, xVectorProjY)
            draw_line(image[:, :, 1], projX, projY, yVectorProjX, yVectorProjY)
            draw_line(image[:, :, 2], projX, projY, zVectorProjX, zVectorProjY)
        return image, (projX, projY)

    def get_frame(self, time, timeWindow, **kwargs):
        allData = self.__data
        if allData is None:
            print('Warning: data is not set')
            return np.zeros((1, 1), dtype=np.uint8)  # This should not happen
        image = np.zeros((self.renderY, self.renderX, 3), dtype=np.uint8)
        # Put a grey box around the edge of the image
        image[0, :, :] = 128
        image[-1, :, :] = 128
        image[:, 0, :] = 128
        image[:, -1, :] = 128
        # Put a grey crosshair in the centre of the image
        rY = self.renderY
        rX = self.renderX
        chp = 20  # Cross Hair Proportion for following expression
        image[int(rY / 2 - rY / chp): int(rY / 2 + rY / chp), int(rX / 2), :] = 128
        image[int(rY / 2), int(rX / 2 - rX / chp): int(rX / 2 + rX / chp), :] = 128
        for dataKey in allData.keys():
            data = allData[dataKey]
            # Note, for the following, this follows library function pose6qInterp, but is broken out here, because of slight behavioural differences.
            idxPre = np.searchsorted(data['ts'], time, side='right') - 1
            timePre = data['ts'][idxPre]
            if timePre == time:
                # In this edge-case of desired time == timestamp, there is no need 
                # to interpolate 
                point = data['point'][idxPre, :]
                rotation = data['rotation'][idxPre, :]
            elif idxPre < 0 or (idxPre >= len(data['ts']) - 1):
                # In this edge-case of the time at the beginning or end, 
                # don't show any pose
                point = None
                rotation = None
            else:
                if kwargs.get('interpolate', True):
                    timePost = data['ts'][idxPre + 1]
                    qPre = data['rotation'][idxPre, :]
                    qPost = data['rotation'][idxPre + 1, :]
                    timeRel = (time - timePre) / (timePost - timePre)
                    rotation = slerp(qPre, qPost, timeRel)
                    locPre = data['point'][idxPre, :]
                    locPost = data['point'][idxPre + 1, :]
                    point = locPre * (1 - timeRel) + locPost * timeRel
                    timeDist = min(time - timePre, timePost - time)
                    if timeDist > timeWindow / 2:
                        # Warn the viewer that this interpolation is 
                        # based on data beyond the timeWindow
                        image[:30, :30, 0] = 255  # TODO: Hardcoded
                else:  # No interpolation, so just choose the sample which is nearest in time
                    poseIdx = findNearest(data['ts'], time)
                    point = data['point'][poseIdx, :]
                    rotation = data['rotation'][poseIdx, :]
            image, coords = self.project_pose(point, rotation, image, **kwargs)
            if kwargs.get('label_multiple_bodies', True) and coords is not None:
                cv2.putText(
                    image,  # numpy array on which text is written
                    dataKey,  # text
                    coords,  # position at which writing has to start
                    cv2.FONT_HERSHEY_SIMPLEX,  # font family
                    0.2,  # font size
                    (255, 255, 255, 255),  # font color
                    1)  # font stroke

        # Allow for arbitrary post-production on image with a callback
        # TODO: as this is boilerplate, it could be pushed into pie syntax ...
        if kwargs.get('callback', None) is not None:
            kwargs['image'] = image
            image = kwargs['callback'](**kwargs)
        return image

    def get_dims(self):
        return self.renderX, self.renderY

    def get_colorfmt(self):
        return 'rgb'

    def get_settings(self):
        settings = {'interpolate': {'type': 'boolean',
                                    'default': True
                                    },
                    'perspective': {'type': 'boolean',
                                    'default': True
                                    }}
        return settings



# -*- coding: utf-8 -*-
'''
Copyright (C) 2019 Event-driven Perception for Robotics
Authors: Sim Bamford
This program is free software: you can redistribute it and/or modify it under 
the terms of the GNU General Public License as published by the Free Software 
Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY 
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
PARTICULAR PURPOSE.  See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with 
this program. If not, see <https://www.gnu.org/licenses/>.

Intended as part of bimvee (Batch Import, Manipulation, Visualisation and Export of Events etc)
plotDvsSpaceTime takes 'inDict' - a dict containing an imported ae file, 
as created by importAe, crops it spatially and temporally according to any kwargs
such as maxTim etc
and creates a point cloud viewer containing 

...
'''

import numpy as np
import pptk

from .split import cropSpaceTime

def plotDvsSpaceTime(inDicts, **kwargs):
    if isinstance(inDicts, list):
        for inDict in inDicts:
            plotDvsSpaceTime(inDict, **kwargs)
        return
    else:
        inDict = inDicts
    if not isinstance(inDict, dict):
        return
    if 'ts' not in inDict:
        #title = kwargs.pop('title', '')
        if 'info' in inDict and isinstance(inDict, dict):
            fileName = inDict['info'].get('filePathOrName')
            if fileName is not None:
                print('plotDvsContrast was called for file ' + fileName)
                #title = (title + ' ' + fileName).lstrip()
        for key in inDict.keys():
        #    kwargs['title'] = (title + ' ' + key).lstrip()
            plotDvsSpaceTime(inDict[key], **kwargs)
        return
    # From this point onwards, it's a data-type container
    if 'pol' not in inDict:
        return
    # From this point onwards, it's a dvs container        

    inDict = cropSpaceTime(inDict, **kwargs)
    
    # scale x and y to match time range
    timeRange = inDict['ts'][-1] - inDict['ts'][0]
    spatialDim = max(max(inDict['x']), max(inDict['y']))
    scalingFactor = timeRange / spatialDim
    events = np.concatenate((inDict['x'][:, np.newaxis] * scalingFactor, 
                             inDict['y'][:, np.newaxis] * scalingFactor, 
                             inDict['ts'][:, np.newaxis]), axis=1)
    pptkViewer = pptk.viewer(events) 
    return pptkViewer



# -*- coding: utf-8 -*-
"""
Copyright (C) 2019 Event-driven Perception for Robotics
Authors: Massimiliano Iacono
This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with
this program. If not, see <https://www.gnu.org/licenses/>.

Part of bimvee (Batch Import, Manipulation, Visualisation and Export of Events etc)
importIitNumpy imports timestamped address-event data, given a path to a .npy file.
It serves the needs of a specific data format developed at IIT involving events, 
frames and bounding boxes.
If available at the predetermined locations, relative to the npy file additional 
data may also be imported, such as frames and bounding boxes used to label the 
data. At the moment frames and boxes are stored in a data.log file using the 
YARP format (see importIitYarp as an example). TODO: generalize data structure

The output is a dictionary containing:
    - info
    - data
The exact contents varies according to the file type import but in general:
    info: this is a dict which starts life as the kwargs passed in, and is
    augmented and modified by the actual contents of the file. It may include
    any informational fields in file headers. Minimally, it will contain:
        - filePathAndName
        - fileFormat
    data: this is a list of dicts, one for each sensor or "channel" which has
    been imported. Bear in mind that sub-functions may optionally split or join
    channels. Within each dict, there is a field for each type of data contained.
    A file for example may contain data from a several sensors, but a single sensor
    may produce polarity events ("pol"), aps samples ("aps"), imu samples etc.
    Within each of these fields, there is a dict, generally containing fields for
    each data column, so in the case of pol events, there are 4-5 fields:
        - ts
        - x
        - y
        - pol
        - optionally ch (channel)
        each containing a numpy array of the appropriate type for the data
        contained, where all these arrays will be of the same length.
    Similarly bounding boxes are saved in the channel in a separate dictionary containing the following fields:
        - ts
        - minY
        - minX
        - maxY
        - maxX
"""

import numpy as np
import os

def importIitNumpy(filePathOrName, **kwargs):
    outDict = {
        'info': kwargs,
        'data': {}
    }
    outDict['info']['filePathOrName'] = filePathOrName

    # Importing events
    events = np.load(filePathOrName)

    outDict['data']['labelledEvents'] = {}
    outDict['data']['labelledEvents']['dvs'] = {}

    tsOffset = events[0, 0]
    ts_to_sync = [events[:, 0]]

    outDict['data']['labelledEvents']['dvs']['ts'] = events[:, 0]
    outDict['data']['labelledEvents']['dvs']['x'] = events[:, 1].astype(int)
    outDict['data']['labelledEvents']['dvs']['y'] = events[:, 2].astype(int)
    outDict['data']['labelledEvents']['dvs']['pol'] = events[:, 3].astype(bool)
    outDict['data']['labelledEvents']['dvs']['dimX'] = 304
    outDict['data']['labelledEvents']['dvs']['dimY'] = 240


    # Importing bounding boxes for events
    gt_filename = os.path.join(os.path.dirname(filePathOrName), 'boxes.npy')

    if os.path.exists(gt_filename):
        b_boxes = np.load(gt_filename)
        outDict['data']['labelledEvents']['boundingBoxes'] = {}

        tsOffset = min(tsOffset, b_boxes[0, 0])
        ts_to_sync.append(b_boxes[:, 0])
        outDict['data']['labelledEvents']['boundingBoxes']['ts'] = b_boxes[:, 0]
        outDict['data']['labelledEvents']['boundingBoxes']['minY'] = b_boxes[:, 1]
        outDict['data']['labelledEvents']['boundingBoxes']['minX'] = b_boxes[:, 2]
        outDict['data']['labelledEvents']['boundingBoxes']['maxY'] = b_boxes[:, 3]
        outDict['data']['labelledEvents']['boundingBoxes']['maxX'] = b_boxes[:, 4]
        outDict['data']['labelledEvents']['boundingBoxes']['label'] = b_boxes[:, 5]

    # Importing frames
    framesPath = (os.path.join(os.path.dirname(filePathOrName), '../processed/frames_left')) # TODO make path argument
    if os.path.exists(framesPath):
        import re
        from imageio import imread
        pattern = re.compile('\d+ (\d+\.\d+) (.+\.\w+) \[rgb\]')
        with open(os.path.join(framesPath, 'data.log')) as f:
            content = f.read()
            found = np.array(pattern.findall(content))

        outDict['data']['labelledFrames'] = {}
        outDict['data']['labelledFrames']['frame'] = {}

        frames_ts = found[:, 0].astype(float)
        tsOffset = min(tsOffset, frames_ts[0])
        ts_to_sync.append(frames_ts)

        outDict['data']['labelledFrames']['frame']['ts'] = frames_ts
        outDict['data']['labelledFrames']['frame']['frames'] = [imread(os.path.join(framesPath, x)) for x in found[:, 1]]


    # Importing Bounding Boxes for frames
    framesPath = (os.path.join(os.path.dirname(filePathOrName), '../processed/gt_left')) # TODO make path argument
    if os.path.exists(framesPath):
        import re
        pattern = re.compile('\d+ (\d+\.\d+) (\d+\.\w+) (.*)') # TODO specific to yarp data format
        pattern2 = re.compile('\( (\d+ \d+ \d+ \d+) \) (\d+) (\d+\.\d+)')
        with open(os.path.join(framesPath, 'data.log')) as f:
            content = f.read()
            found = np.array(pattern.findall(content))

        boxes = []
        boxes_ts = []
        labels = []

        for ts, frame_name, bbs in found:
            for box, label, _ in pattern2.findall(bbs): # TODO third element is confidence score. not used at the moment
                boxes.append(box.split(' '))
                labels.append(label)
                boxes_ts.append(ts)

        boxes_ts = np.array(boxes_ts).astype(float)
        labels = np.array(labels).astype(int)
        boxes = np.array(boxes).astype(int)

        tsOffset = min(tsOffset, boxes_ts[0])
        ts_to_sync.append(boxes_ts)
        outDict['data']['labelledFrames']['boundingBoxes'] = {}
        outDict['data']['labelledFrames']['boundingBoxes']['ts'] = boxes_ts
        outDict['data']['labelledFrames']['boundingBoxes']['minY'] = boxes[:, 0]
        outDict['data']['labelledFrames']['boundingBoxes']['minX'] = boxes[:, 1]
        outDict['data']['labelledFrames']['boundingBoxes']['maxY'] = boxes[:, 2]
        outDict['data']['labelledFrames']['boundingBoxes']['maxX'] = boxes[:, 3]
        outDict['data']['labelledFrames']['boundingBoxes']['label'] = labels

    for x in ts_to_sync:
        x -= tsOffset

    return outDict


# -*- coding: utf-8 -*-
"""
Copyright (C) 2020 Event-driven Perception for Robotics
Author: Sim Bamford

This program is free software: you can redistribute it and/or modify it under 
the terms of the GNU General Public License as published by the Free Software 
Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY 
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
PARTICULAR PURPOSE.  See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with 
this program. If not, see <https://www.gnu.org/licenses/>.

Intended as part of bimvee (Batch Import, Manipulation, Visualisation and Export of Events etc)
Low-level gemeotry functions, 
for example, for manipulating poses, orientations, projections etc
"""

import numpy as np

# local imports
from .timestamps import sortDataTypeDictByTime

# Can accept an existing matrix, which should be min 3x3; 
#if it creates a matrix it makes it 4x4
def quat2RotM(quat, M=None):
    if M is None: 
        M = np.zeros((4, 4))
        M[3, 3] = 1
    elif M.shape[0] == 3:
        M[:, 3] = 0
        M[3, :] = 0
        M[3, 3] = 1
    w = quat[0]
    x = quat[1]
    y = quat[2]
    z = quat[3]
    M[0, 0] = 1 - 2*y**2 - 2*z**2
    M[0, 1] = 2*x*y - 2*z*w
    M[0, 2] = 2*x*z + 2*y*w
    M[1, 0] = 2*x*y + 2*z*w
    M[1, 1] = 1 - 2*x**2 - 2*z**2
    M[1, 2] = 2*y*z - 2*x*w
    M[2, 0] = 2*x*z - 2*y*w
    M[2, 1] = 2*y*z + 2*x*w
    M[2, 2] = 1 - 2*x**2 - 2*y**2
    return M

# Spherical linear interpolation, adapted from https://en.wikipedia.org/wiki/Slerp
DOT_THRESHOLD = 0.9995
def slerp(q1, q2, time_relative):
    dot = np.sum(q1 * q2)
    if dot < 0.0:
        q2 = -q2
        dot = -dot
    if dot > DOT_THRESHOLD:
        result = q1 + time_relative * (q2 - q1)
        return (result.T / np.linalg.norm(result)).T
    theta_0 = np.arccos(dot)
    sin_theta_0 = np.sin(theta_0)
    theta = theta_0 * time_relative
    sin_theta = np.sin(theta)
    s0 = np.cos(theta) - dot * sin_theta / sin_theta_0
    s1 = sin_theta / sin_theta_0
    return (s0 * q1) + (s1 * q2)

''' 
expects pose dict in the form: {'ts': 1d np.array of float timestamps,
                                'point': 2d array float of positions [x, y, z], 
                                'rotation': 2d array float of quaternions [rw, rx, ry, rz]
                                (i.e. 6dof with rotation as quaternion)}
Two modes of operation:
If time is not None, then returns the interpolated pose at that time -
    returns (point, rotation) tuple, being np.array 1d x 3 and 4 respectively, float,    which is interpolated pose;
Else if maxPeriod is not None, then it returns the entire pose dict,
    but with additional points necessary to ensure that time between samples 
    never exceeds maxPeriod
'''
def pose6qInterp(poseDict, time=None, maxPeriod=None):
    ts = poseDict['ts']
    points = poseDict['point']
    rotations = poseDict['rotation']
    if time is not None:
        idxPre = np.searchsorted(ts, time, side='right') - 1
        timePre = ts[idxPre]
        if timePre == time:
            # In this edge-case of desired time == timestamp, there is no need 
            # to interpolate 
            return (points[idxPre, :], rotations[idxPre, :])
        if idxPre < 0:
            return (points[0, :], rotations[0, :])
        if idxPre >= len(poseDict['ts']):
            return (points[-1, :], rotations[-1, :])
        timePost = ts[idxPre + 1]
        rotPre = rotations[idxPre, :]
        rotPost = rotations[idxPre + 1, :]
        timeRel = (time - timePre) / (timePost - timePre)
        rotOut = slerp(rotPre, rotPost, timeRel)
        pointPre = points[idxPre, :] 
        pointPost = points[idxPre + 1, :]
        pointOut = pointPre * (1-timeRel) + pointPost * timeRel
        return (pointOut, rotOut)
    elif maxPeriod is not None:
        firstTime = ts[0]
        lastTime = ts[-1]
        proposedAdditionalTimes = np.arange(firstTime, lastTime, maxPeriod)
        prevIds = np.searchsorted(ts, proposedAdditionalTimes, side='right') - 1
        distPre = proposedAdditionalTimes - ts[prevIds]
        distPost = ts[prevIds + 1] - proposedAdditionalTimes
        dist = distPre + distPost
        keepAdditional = dist > maxPeriod
        additionalTimes = proposedAdditionalTimes[keepAdditional]
        additionalPoses = [pose6qInterp(poseDict, time=additionalTime) 
                            for additionalTime in additionalTimes] 
        additionalPoints, additionalRotations = zip(*additionalPoses)
        ts = np.concatenate((ts, additionalTimes))
        points = np.concatenate((points, np.array(additionalPoints)))
        rotations = np.concatenate((rotations, np.array(additionalRotations)))
        poseDict['ts'] = ts
        poseDict['point'] = points
        poseDict['rotation'] = rotations
        poseDict = sortDataTypeDictByTime(poseDict)
        return poseDict

def quaternionProduct(q1, q2):
    return np.array([ 
    q1[0] * q2[0] - q1[1] * q2[1] - q1[2] * q2[2] - q1[3] * q2[3],
    q1[0] * q2[1] + q1[1] * q2[0] + q1[2] * q2[3] - q1[3] * q2[2],
    q1[0] * q2[2] - q1[1] * q2[3] + q1[2] * q2[0] + q1[3] * q2[1],
    q1[0] * q2[3] + q1[1] * q2[2] - q1[2] * q2[1] + q1[3] * q2[0] ])
    ''' alternative definition - less efficient
    w1 = q1[0]
    v1 = q1[1:4]
    w2 = q2[0]
    v2 = q2[1:4]
    wOut = w1*w2 - np.dot(v1.T, v2)
    vOut = w1*v2 + w2*v1 + np.cross(v1, v2)
    qOut = np.zeros_like(q1)
    qOut[0] = wOut
    qOut[1:4] = vOut
    return qOut
    '''

# Legacy name for Quaternion multiplication
def combineTwoQuaternions(q1, q2):
    return quaternionProduct(q1, q2)

def quaternionConjugate(q):
    return np.array([q[0], -q[1], -q[2], -q[3]])

def quaternionInverse(q):
    return quaternionConjugate(q) / np.sum(q ** 2)

def angleBetweenTwoQuaternions(q1, q2):
    # returns minimal angle in radians between two unit quaternions, following:
    # https://www.researchgate.net/post/How_do_I_calculate_the_smallest_angle_between_two_quaternions    
    qP = quaternionProduct(q1, quaternionInverse(q2))
    normV = np.linalg.norm(qP[1:])
    return 2 * np.arcsin(normV)

'''
following https://github.com/KieranWynn/pyquaternion/blob/master/pyquaternion/quaternion.py
'''
def axisAngle2Quat(axis, angle):
    mag_sq = np.dot(axis, axis)
    if mag_sq == 0.0:
        raise ZeroDivisionError("Provided rotation axis has no length")
    # Ensure axis is in unit vector form
    if (abs(1.0 - mag_sq) > 1e-12):
        axis = axis / np.sqrt(mag_sq)
    theta = angle / 2.0
    r = np.cos(theta)
    i = axis * np.sin(theta)
    quat = np.concatenate((np.ones((1)) * r, i))
    quat = quat / np.linalg.norm(quat)
    return quat

'''
following: https://answers.unity.com/questions/1266985/convert-rotation-vector-or-matrix-to-quaternion.html
Receives a rotation vector and returns a quaternion
'''
def rotVToQuat(rotV):
    theta = np.linalg.norm(rotV)
    quat = axisAngle2Quat(rotV[:, 0], theta)
    return quat
    
'''
Expects 
    - poseDict in bimvee form {'ts', 'point', 'rotation' as above}
    - translation as np array of x,y,z
    - rotation as a np array of w,x,y,z (quaternion)
If either translation or rotation are passed, these are applied to all poses 
in the poseDict.
The rotations are defined wrt local axes, unlike the translations.
Returns a copy of the poseDict, rotated
'''    
def transformPoses(poseDict, translation=None, rotation=None):
    # Create a copy of the input array - use the same contents
    outDict = {}
    for key in poseDict.keys():
        outDict[key] = poseDict[key]
    if translation is not None:
        outDict['point'] = poseDict['point'] + translation
    if rotation is not None:
        for idx in range(outDict['rotation'].shape[0]): # TODO: this could be matricised
            outDict['rotation'][idx, :] = \
                combineTwoQuaternions(outDict['rotation'][idx, :], rotation)
    return outDict
        

# The following adapted from https://github.com/christophhagen/averaging-quaternions
# Note that the signs of the output quaternion can be reversed, 
# since q and -q describe the same orientation.
# w is an optional weight array, which must have the same number of elements 
# as the number of quaternions
def averageOfQuaternions(allQ, w=None):
    # Number of quaternions to average
    M = allQ.shape[0]
    A = np.zeros((4,4), dtype = float)
    if w is None:
        w = np.ones(M,)
    weightSum = 0
    for i in range(0, M):
        q = allQ[i, :]
        # multiply q with its transposed version q' and add A
        A = w[i] * np.outer(q,q) + A
        weightSum += w[i]
    # scale
    A = (1.0 / M) * A
    # compute eigenvalues and -vectors
    eigenValues, eigenVectors = np.linalg.eig(A)
    # Sort by largest eigenvalue
    eigenVectors = eigenVectors[:,eigenValues.argsort()[::-1]]
    # return the real part of the largest eigenvector (has only real part)
    return eigenVectors[:,0]



import numpy as np


def importBoundingBoxes(**kwargs):
    bboxes = np.loadtxt(kwargs.get('filePathOrName'))
    if len(bboxes.shape) == 1:
        bboxes = np.expand_dims(bboxes,0)
    bboxes = bboxes[np.argsort(bboxes[:, 0])]
    outDict = {'ts': bboxes[:, 0],
               'minY': bboxes[:, 1],
               'minX': bboxes[:, 2],
               'maxY': bboxes[:, 3],
               'maxX': bboxes[:, 4],
               'label': bboxes[:, 5],
               }
    return outDict



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Copyright (C) 2020 Event-driven Perception for Robotics
Authors: Sim Bamford
         Aiko Dinale
This program is free software: you can redistribute it and/or modify it under 
the terms of the GNU General Public License as published by the Free Software 
Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY 
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
PARTICULAR PURPOSE.  See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with 
this program. If not, see <https://www.gnu.org/licenses/>.

Intended as part of bimvee (Batch Import, Manipulation, Visualisation and Export of Events etc)
plotImu takes 'inDict' - a dictionary containing imported IMU data 
(or a higher level container, in which attempts to descend and call itself) 
as created by importAe, and plots against time the various dimensions of the 
IMU samples contained. 
"""
import os
import matplotlib.pyplot as plt
import seaborn as sns

sns.set(palette="colorblind")

# Color Palette for Color Blindness
zesty_palette     = ['#F5793A', '#A95AA1', '#85C0F9', '#0F2080']
retro_palette     = ['#601A4A', '#EE442F', '#63ACBE', '#F9F4EC']
corporate_palette = ['#8DB8AD', '#EBE7E0', '#C6D4E1', '#44749D']

#-----------------------------------------------------------------------------------------------------
def plotFlow(flowDict, fig_path=None, fig_name=None, fig_subtitle=None):
    """
    Plot the FLOW events in flowDict against the time. If specified, save the
    generated figure as fig_name.png at the location defined by fig_path.

    Arguments:
        flowDict {dict} -- dictionary of FLOW events as formatted by bimvee from event-driven library

    Keyword Arguments:
        fig_path {string} -- save path for the generated figure (default: {None})
        fig_name {string} -- name of the generated figure (default: {None})
        fig_subtitle {string} -- figure sub-title (default: {None})
    """
    fig = plt.figure(figsize=(16.0, 10.0))
    if isinstance(fig_subtitle, str):
        fig.suptitle("FLOW Events\n" + fig_subtitle, fontsize=20, fontweight='bold')
    else:
        fig.suptitle("FLOW Events", fontsize=20, fontweight='bold')


    ax11 = plt.subplot(2, 1, 1)
    plt.plot(flowDict['ts'], flowDict['vx'], color=retro_palette[0], marker='.', linewidth=0.0)
    plt.xlabel('Time [s]', fontsize=10, fontweight='bold')
    plt.ylabel('Vx [px/s]', fontsize=10, fontweight='bold')
    plt.grid(True)

    plt.subplot(2, 1, 2, sharex=ax11)
    plt.plot(flowDict['ts'], flowDict['vy'], color=retro_palette[1], marker='.', linewidth=0.0)
    plt.xlabel('Time [s]', fontsize=10, fontweight='bold')
    plt.ylabel('Vy [px/s]', fontsize=10, fontweight='bold')
    plt.grid(True)

    fig.tight_layout()
    if isinstance(fig_subtitle, str):
        fig.subplots_adjust(top=0.9)
    else:
        fig.subplots_adjust(top=0.95)

    fig.align_ylabels()

    if isinstance(fig_path, str) and isinstance(fig_name, str):
        plt.savefig(os.path.join(fig_path, fig_name + ".png"), dpi=300, bbox_inches='tight')
        print("Saving " + fig_name + ".png")
        plt.close()
    else:
        plt.show()

#-----------------------------------------------------------------------------------------------------
def plotFlowDistribution(flowDict, fig_path=None, fig_name=None, fig_subtitle=None):
    """
    Plot the distribution of the FLOW events in flowDict and save the generated
    figure as fig_name.png at the location defined by fig_path.

    Arguments:
        flowDict {dict} -- dictionary of FLOW events as formatted by bimvee from event-driven library

    Keyword Arguments:
        fig_path {string} -- save path for the generated figure (default: {None})
        fig_name {string} -- name of the generated figure (default: {None})
        fig_subtitle {string} -- figure sub-title (default: {None})
    """
    fig = plt.figure(figsize=(14.0, 10.0))
    if isinstance(fig_subtitle, str):
        fig.suptitle("FLOW Events Distribution\n" + fig_subtitle, fontsize=20, fontweight='bold')
    else:
        fig.suptitle("FLOW Events Distribution", fontsize=20, fontweight='bold')
    
    plt.subplot(2,1,1)
    sns.distplot(flowDict['vx'], bins=100, color = retro_palette[0])
    plt.xlabel('Vx [px/s]', fontsize=10, fontweight='bold')

    plt.subplot(2,1,2)
    sns.distplot(flowDict['vy'], bins=100, color = retro_palette[1])
    plt.xlabel('Vy [px/s]', fontsize=10, fontweight='bold')

    fig.tight_layout()
    if isinstance(fig_subtitle, str):
        fig.subplots_adjust(top=0.9)
    else:
        fig.subplots_adjust(top=0.95)

    if isinstance(fig_path, str) and isinstance(fig_name, str):
        plt.savefig(os.path.join(fig_path, fig_name + '.png'), dpi=300, bbox_inches='tight')
        print("Saving " + fig_name + ".png")
        plt.close()
    else:
        plt.show()



# -*- coding: utf-8 -*-

"""https://github.com/event-driven-robotics/importRosbag
Copyright (C) 2019 Event-driven Perception for Robotics
Authors: Sim Bamford
This program is free software: you can redistribute it and/or modify it under 
the terms of the GNU General Public License as published by the Free Software 
Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY 
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
PARTICULAR PURPOSE.  See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with 
this program. If not, see <https://www.gnu.org/licenses/>.

Intended as part of bimvee (Batch Import, Manipulation, Visualisation and Export of Events etc)
importRpgDvsRos uses the importRosbag submodule
(original located at https://github.com/event-driven-robotics/importRosbag)
to import a rosbag, containing a selected set of ros msg types.
In particular, this supports the dvs_msgs/EventArray messages defined at:
http://rpg.ifi.uzh.ch/davis_data.html
It also supports some standard ros types:
    geometry_msgs/PoseStamped 
    geometry_msgs/Transform
    geometry_msgs/TransformStamped
    geometry_msgs/TwistStamped
    sensor_msgs/CameraInfo, (calibration)
    sensor_msgs/Image
    sensor_msgs/Imu 
    sensor_msgs/PointCloud2
    tf/tfMessage
Furthermore there is support for:
    esim_msgs/OpticFlow

It returns nested dicts of the form:
{
    info
    data
        channelName
            dvs
                "pol": numpy array of bool
                "x": numpy array of uint16
                "y": numpy array of uint16
                "ts": numpy array of float - seconds (basic format is int with unit increments of 80 ns) 
            frame ...
            imu ...
            etc ...

It optionally reorders the resulting data to support legacy templates (kwarg=template) 
Here follow 2 example templates:
    
template = {
    'left': {
        'dvs': '/davis/left/events',
    }, 'right': {
        'dvs': '/davis/right/events',
        }
    }    
    
template = {
    'ch0': {
        'dvs': '/dvs/events',
        'frame': '/dvs/image_raw',
        'pose6q': '/optitrack/davis',
        'cam': '/dvs/camera_info',
        'imu': '/dvs/imu'
        }
    }    

If a template is supplied, any connections which are not named in the template are not imported but simply listed.
If an empty template is supplied, then the contained types are printed out.
Use this to inspect a new .bag file before defining the import template.
Without any template, the default behaviour is to put any imported data into 
its own channel, named after the topic in the bag.
"""

#%%

import string

# Local imports
from .timestamps import zeroTimestampsForAChannel, rezeroTimestampsForImportedDicts
from .importRosbag.importRosbag.importRosbag import importRosbag
    
def bimveeTypeForRosbagType(rosbagType):
    rosbagType = rosbagType.replace('/','_')
    if rosbagType == 'dvs_msgs_EventArray': return 'dvs'
    if rosbagType == 'esim_msgs_OpticFlow': return 'flowMap'
    if rosbagType == 'geometry_msgs_PoseStamped': return 'pose6q'
    if rosbagType == 'geometry_msgs_Transform': return 'pose6q'
    if rosbagType == 'geometry_msgs_TransformStamped': return 'pose6q'
    if rosbagType == 'geometry_msgs_TwistStamped': return 'twist'
    if rosbagType == 'sensor_msgs_CameraInfo': return 'cam'
    if rosbagType == 'sensor_msgs_Image': return 'frame'
    if rosbagType == 'sensor_msgs_Imu': return 'imu'
    if rosbagType == 'sensor_msgs_PointCloud2': return 'point3'
    if rosbagType == 'sensor_msgs_Illuminance': return 'illuminance'
    if rosbagType == 'nav_msgs_Odometry': return 'pose6q'
    if rosbagType == 'tf_tfMessage': return 'pose6q'
    if rosbagType == 'vicon_Subject': return 'pose6q'
    return None

def importRpgDvsRos(filePathOrName, **kwargs):    
    template = kwargs.get('template')
    if template == {}: # Just list contents of bag without importing
        topics = importRosbag(filePathOrName=filePathOrName, listTopics=True, **kwargs)
        return
    topics = importRosbag(filePathOrName=filePathOrName, **kwargs)
    outDict = {
        'info': kwargs,
        'data': {}
            }
    outDict['info']['filePathOrName'] = filePathOrName
    if template is None:
        for topicLabel in topics.keys():
            rosbagType = topics[topicLabel].pop('rosbagType')
            bimveeType = bimveeTypeForRosbagType(rosbagType)
            if bimveeType is None:
                print('Actually, ' + topicLabel + ' has not been imported, because the rosbag message type ' + rosbagType + ' has not been recognised.')
            else:
                outDict['data'][topicLabel] = {bimveeType: topics[topicLabel]}
    else:
        # If we get to here then there is a template to parse
        # The template becomes the data branch of the importedDict
        for channelKey in template.keys():
            channelKeyStripped = str(channelKey).translate(str.maketrans('', '', string.punctuation))
            outDict['data'][channelKeyStripped] = {}
            for dataType in template[channelKey]:
                topicLabel = template[channelKey][dataType]
                topic = topics.pop(topicLabel)
                rosbagType = topic.pop('rosbagType')
                bimveeType = bimveeTypeForRosbagType(rosbagType)
                if bimveeType != dataType:
                    print('dataType "', dataType, '" not correctly defined for topic: "', topicLabel, '"')
                else:
                    outDict['data'][channelKeyStripped][dataType] = topic
    # Post processing
    if kwargs.get('zeroTime', kwargs.get('zeroTimestamps', True)):
        # Optional: start the timestamps at zero for the first event
        # This is done collectively for all the concurrent imports
        for channelKey in outDict['data'].keys():
            zeroTimestampsForAChannel(outDict['data'][channelKey])
        # jointly rezero for all channels
        rezeroTimestampsForImportedDicts(outDict)
    return outDict



# -*- coding: utf-8 -*-

"""
Copyright (C) 2020 Event-driven Perception for Robotics
Authors: Sim Bamford
         Ander Arriandiaga Laresgoiti

This program is free software: you can redistribute it and/or modify it under 
the terms of the GNU General Public License as published by the Free Software 
Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY 
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
PARTICULAR PURPOSE.  See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with 
this program. If not, see <https://www.gnu.org/licenses/>.

Intended as part of bimvee (Batch Import, Manipulation, Visualisation and Export of Events etc)
importSecDvs opens a .bin file, assumed to contain encoded data from SecDvs gen3.
Returns a dict in this format:
{'info': {},
 'data': {
         ch0: {
               dvs: {
                     'ts': np.array of float in seconds 
                     'x': np.array of np.uint16 in pixels
                     'y': np.array of np.uint16 in pixels
                     'pol': np.array of bool -- 1 = ON event 
                    }}}}
"""

import numpy as np
from tqdm import tqdm 

# Local imports
from .timestamps import unwrapTimestamps, zeroTimestampsForADataType

def importSecDvs(**kwargs):
    filePathOrName = kwargs['filePathOrName']
    print('Attempting to import ' + filePathOrName + ' as secdvs')
    with open(filePathOrName, 'rb') as file:
        data = np.fromfile(file, dtype='>u4')

    print('Clipping any rows without column info...')
    # Hunt for the first col iteratively and cut data before this
    for idx, word in enumerate(data):
        if word & 0x4000000:
            break
    data = data[idx:]

    print('Building indices for word types  ...')
    isTs = (data & 0x08000000).astype(bool) # Reference timestamp, i.e. tsMsb
    isCol = (data & 0x4000000).astype(bool) # X and tsLsb
    isRow = (data & 0x80000000).astype(bool) # y and pol

    print('Handling col data ...')
    # create an index which points data back to col
    numCol = np.count_nonzero(isCol)
    colIdsRange = np.arange(numCol)
    colIdsSparse = np.where(isCol)[0]
    colIdsNext = np.append(colIdsSparse[1:], len(data))
    colIdx = np.zeros_like(data)
    for rangeIdx, firstIdx, nextIdx in zip(colIdsRange, colIdsSparse, colIdsNext):
        colIdx[firstIdx:nextIdx] = rangeIdx
    # now we can isolate col
    col = data[isCol]
    # convert col data
    xByCol = (col & 0x000003FF).astype(np.uint16) # column Address
    tsSmallByCol = (col & 0x1FF800) >> 11
	# create col data vectors which match data
    x = xByCol[colIdx]
    tsSmall = tsSmallByCol[colIdx]
    
    print('Handling timestamp data ...')

    # from data extract the "start col" flag
    isStartCol = (data & 0x200000).astype(bool) & isCol
    # now, for each startCol, we want to search backwards through data for a ts    
    # To do this, we find the "data" idx of start col, we create a set of tsIds
    # and then we search tsIds for each dataIdx    
    tsIdsSparse = np.where(isTs)[0]
    # Actually find the tsMsb at this point, to avoid more indexing
    tsMsbSparse = (data[isTs] & 0x003FFFFF) << 10
    tsMsbSparse = np.insert(tsMsbSparse, 0, tsMsbSparse[0] - (1 << 10))
    # create an index which points data back to startCols
    startColIdsSparse = np.where(isStartCol)[0]
    # Search tsIds for each dataIdx
    tsForStartCol = np.zeros_like(startColIdsSparse)
    for idx, startColIdx in enumerate(startColIdsSparse):
        tsForStartCol[idx] = tsMsbSparse[np.searchsorted(tsIdsSparse, startColIdx)]

    # Now we have the ts(Large) for each startCol event. Now create a tsLarge
    # array which matches data; do this in just the same way as above for
    # tsSmall given col and colIdx
    # create an index which points data back to col
    numStartCol = np.count_nonzero(isStartCol)
    startColIdsRange = np.arange(numStartCol)
    startColIdsSparse = np.where(isStartCol)[0]
    startColIdsNext = np.append(startColIdsSparse[1:], len(data))
    startColIdx = np.zeros_like(data)
    for rangeIdx, firstIdx, nextIdx in zip(startColIdsRange, startColIdsSparse, startColIdsNext):
        startColIdx[firstIdx:nextIdx] = rangeIdx
    tsLarge = tsForStartCol[startColIdx]

    # Now we have tsLarge and tsSmall aligned by data, 
    # we can sum these to give the full ts
    ts = tsLarge + tsSmall

    print('Handling row data ...')

    # Now we have x and ts for each row in data; 
    # now select these just for group/row data
    x = x[isRow]
    ts = ts[isRow]
    data = data[isRow]

    # A major timewrap is possible, so handle unwrapping before the following 
    # processing, which will mix up the timestamps
    # Noticing that for whatever reason, there can be discontinuities in the 
    # minor timestamps, there is a first call to unwrap timestamps which 
    # uses a wrapTime of 1024 us. This would introduce about 1 ms of delay
    # in the case of an actual major timewrap
    ts = unwrapTimestamps(ts, wrapTime=1024)  
    ts = unwrapTimestamps(ts)  / 1000000 # Convert to seconds in the same step
    
    # Break out addr and pol for each of the two groups
    pol1 = ((data & 0x00010000) >> 16).astype(bool)
    yLarge1 = ((data & 0x00FC0000) >> 15).astype(np.uint16) # grp1Address
    pol2 = ((data & 0x00020000) >> 17).astype(bool)
    grp2Offset = (data & 0x7C000000) >> 23 
    yLarge2 = (grp2Offset + yLarge1).astype(np.uint16)

    # for each of the single bit indices, select events for each of the two groups
    grp1Events = data & 0xFF
    grp2Events = data & 0xFF00
    tsToConcatenate = []
    xToConcatenate = []
    yToConcatenate = []
    polToConcatenate = []
    for idx in tqdm(range(8)):
        #group 1
        grp1Bool = (grp1Events & (2 ** idx)).astype(bool)
        tsToConcatenate.append(ts[grp1Bool])
        xToConcatenate.append(x[grp1Bool])
        yToConcatenate.append(yLarge1[grp1Bool] + idx)
        polToConcatenate.append(pol1[grp1Bool])
        # group 2        
        grp2Bool = (grp2Events & (2 ** (idx + 8))).astype(bool)
        tsToConcatenate.append(ts[grp2Bool])
        xToConcatenate.append(x[grp2Bool])
        yToConcatenate.append(yLarge2[grp2Bool] + idx)
        polToConcatenate.append(pol2[grp2Bool])

    print('Post-processing steps ...')

    # Concatenate the resulting arrays
    ts = np.concatenate(tsToConcatenate)
    x = np.concatenate(xToConcatenate)
    y = np.concatenate(yToConcatenate)
    pol = np.concatenate(polToConcatenate)
    
    # The above selection strategy mixed up the timestamps, so sort the events by ts
    
    ids = np.argsort(ts)
    
    ts = ts[ids]
    x = x[ids]
    y = y[ids]
    pol = pol[ids]
    
    # Invert polarity, in line with library-wide definition
    pol = ~pol
    
    dvsDict = {'ts': ts,
               'x': x,
               'y': y,
               'pol': pol,
               }

    if kwargs.get('zeroTime', kwargs.get('zeroTimestamps', True)): 
        zeroTimestampsForADataType(dvsDict)
    outDict = {
    'info': {'filePathOrName':filePathOrName,
        'fileFormat': 'secdvs'},
    'data': {
        'ch0': {
            'dvs': dvsDict
            }
        }
    }
    print('Done.')
            
    return outDict



# -*- coding: utf-8 -*-
"""
Copyright (C) 2019 Event-driven Perception for Robotics
Authors: Sim Bamford
        Suman Ghosh
This program is free software: you can redistribute it and/or modify it under 
the terms of the GNU General Public License as published by the Free Software 
Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY 
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
PARTICULAR PURPOSE.  See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with 
this program. If not, see <https://www.gnu.org/licenses/>.

Intended as part of bimvee (Batch Import, Manipulation, Visualisation and Export of Events etc)
plotPose takes 'inDict' - a dictionary containing imported pose data 
(or a higher level container, in which attempts to descend and call itself) 
as created by importAe, and plots against time the various dimensions of the 
samples contained. Each dimension is plotted against time, 
and each channel has a separate figure

plotPose3d works differently; it takes a dict with multiple channels, rejects
time and rotation info and plots the point data in a 3d space, with one set of
points for each channel of dataType pose6q or point3, in a different colour. 

plotPoseMultiChannel uses the same inputs as plotPose3d to put multiple channels
of pose or point data on the same graph.

Also handles point3 type, which contains 3d points
"""
import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits import mplot3d
import itertools

# local imports
from .split import cropTime
from bimvee.pose import canoniseQuaternions, makeQuaternionsContinuous

def plotPose(inDicts, **kwargs):
    if isinstance(inDicts, list):
        for inDict in inDicts:
            plotPose(inDict, **kwargs)
        return
    else:
        inDict = inDicts
    if not isinstance(inDict, dict):
        return
    if 'ts' not in inDict:
        title = kwargs.pop('title', '')
        if 'info' in inDict and isinstance(inDict, dict):
            fileName = inDict['info'].get('filePathOrName')
            if fileName is not None:
                print('plotPose was called for file ' + fileName)
                title = (title + ' ' + fileName).lstrip()
        for key in inDict.keys():
            kwargs['title'] = (title + ' ' + key).lstrip()
            plotPose(inDict[key], **kwargs)
        return
    # From this point onwards, it's a data-type container
    if 'point' not in inDict:
        return
    # From this point onwards, it's a pose or point data-type container    

    ts = inDict['ts']
    minTime = kwargs.get('minTime', kwargs.get('startTime', kwargs.get('beginTime', ts.min())))
    maxTime = kwargs.get('maxTime', kwargs.get('stopTime', kwargs.get('endTime', ts.max())))
    if minTime > ts.min() or maxTime < ts.max():
        inDict = cropTime(inDict, minTime=minTime, maxTime=maxTime, zeroTime=False)
        ts = inDict['ts']

    axesR = kwargs.get('axesR')
    axesT = kwargs.get('axesT')
    label = kwargs.get('label', '')
    lineStyle = kwargs.get('lineStyle', '-')
    if 'rotation' in inDict:
        if axesR is None:
            fig, allAxes = plt.subplots(2, 1)
            fig.suptitle(kwargs.get('title', ''))
            axesT = allAxes[0]
            axesR = allAxes[1]
        if kwargs.get('canonise', False):
            inDict = canoniseQuaternions(inDict)
        if kwargs.get('makeContinuous', True):
            inDict = makeQuaternionsContinuous(inDict)
        rotation = inDict['rotation']
        axesR.plot(ts, rotation[:, 0], 'k' + lineStyle)
        axesR.plot(ts, rotation[:, 1], 'r' + lineStyle)
        axesR.plot(ts, rotation[:, 2], 'g' + lineStyle)
        axesR.plot(ts, rotation[:, 3], 'b' + lineStyle)
        try:
            existingLegend = [txt._text for txt in axesR.get_legend().get_texts()]
        except AttributeError:
            existingLegend = []
        axesR.legend(existingLegend + 
                     [' '.join([label, 'r_w']), 
                      ' '.join([label, 'r_x']),
                      ' '.join([label, 'r_y']),
                      ' '.join([label, 'r_z']),])
        axesR.set_xlabel('Time (s)')
        axesR.set_ylabel('Quaternion components')
        axesR.set_ylim([-1, 1])
    if 'point' in inDict:
        if axesT is None:
            fig, axesT = plt.subplots()
        if kwargs.get('zeroT', False):
            point = np.copy(inDict['point'])
            firstPoint = inDict['point'][0, :]
            point = point - firstPoint
        else:
            point = inDict['point']
        axesT.plot(ts, point[:, 0], 'r' + lineStyle)
        axesT.plot(ts, point[:, 1], 'g' + lineStyle)
        axesT.plot(ts, point[:, 2], 'b' + lineStyle)
        try:
            existingLegend = [txt._text for txt in axesT.get_legend().get_texts()]
        except AttributeError:
            existingLegend = []
        axesT.legend(existingLegend + 
                     [' '.join([label, 'x']),
                      ' '.join([label, 'y']),
                      ' '.join([label, 'z']),])
        axesT.set_xlabel('Time (s)')
        axesT.set_ylabel('Coords (m)')
    callback = kwargs.get('callback')
    if callback is not None:
        del kwargs['callback']
        kwargs['axesT'] = axesT
        kwargs['axesR'] = axesR
        kwargs['minTime'] = minTime
        kwargs['maxTime'] = maxTime
        callback(**kwargs)

def plotPoseMultiChannel(inDict, **kwargs):
    if 'data' in inDict:
        inDict = inDict['data']
    axesR = kwargs.get('axesR')
    if axesR is None:
        fig, allAxes = plt.subplots(2, 1)
        fig.suptitle(kwargs.get('title', ''))
        kwargs['axesT'] = allAxes[0]
        kwargs['axesR'] = allAxes[1]
    bodyIds = kwargs.get('bodyIds', inDict.keys())
    include = kwargs.get('include', [])
    exclude = kwargs.get('exclude', [])
    lineStyle = itertools.cycle(('-', '--', '-.', ':'))
    for name in bodyIds:
        select_body = all([(inc in name) for inc in include]) and all([not (exc in name) for exc in exclude])
        if select_body:  # modify this line to plot whichever markers you want
            try:
                subDataDict = inDict[name]['pose6q']
            except KeyError:
                try:
                    subDataDict = inDict[name]['point3']
                except KeyError:
                    continue
            kwargs['label'] = name
            kwargs['lineStyle'] = next(lineStyle)
            plotPose(subDataDict, **kwargs)
            

"""
Plot the 3D trajectory of a body or marker for the entire duration of recording

To select which bodies to plot using bodyID: 
 - look at the bodyIDs parsed using importVicon
 - pass a list of strings present in the bodyID of choice, through the parameter include
 - pass a list of strings that should be absent from the bodyID of choice, through the parameter exclude
"""

# TODO: this function could be built up to become the core of pose3dMultiChannel
def plotPose3d(points, **kwargs):
    ax = plt.axes(projection='3d')
    X = points[0, :]
    Y = points[1, :]
    Z = points[2, :]
    ax.scatter3D(X, Y, Z, marker='o', s=0.5)
    ax.set_xlabel('X')
    ax.set_ylabel('Y')
    ax.set_zlabel('Z')
    plt.show()

# legacy name for the above function
def plotPoints(points, **kwargs):
    plotPose3d(points, **kwargs)

def plotPose3dMultiChannel(dataDict, **kwargs):
    if 'data' in dataDict:
        dataDict = dataDict['data']
    ax = kwargs.get('ax')
    if ax is None:
        fig, ax = plt.subplots(projection='3d')
        kwargs['ax'] = ax
    bodyIds = kwargs.get('bodyIds', dataDict.keys())
    include = kwargs.get('include', [])
    exclude = kwargs.get('exclude', [])
    handles = []
    for name in bodyIds:
        select_body = all([(inc in name) for inc in include]) and all([not (exc in name) for exc in exclude])
        if select_body:  # modify this line to plot whichever markers you want
            marker_pose = dataDict[name]['pose6q']['point']
            X = marker_pose[:, 0]
            Y = marker_pose[:, 1]
            Z = marker_pose[:, 2]
            handles.append(ax.scatter3D(X, Y, Z, label=name))
            
            # Create cubic bounding box to simulate equal aspect ratio
            max_range = np.array([X.max() - X.min(), Y.max() - Y.min(), Z.max() - Z.min()]).max()
            Xb = 0.5 * max_range * np.mgrid[-1:2:2, -1:2:2, -1:2:2][0].flatten() + 0.5 * (X.max() + X.min())
            Yb = 0.5 * max_range * np.mgrid[-1:2:2, -1:2:2, -1:2:2][1].flatten() + 0.5 * (Y.max() + Y.min())
            Zb = 0.5 * max_range * np.mgrid[-1:2:2, -1:2:2, -1:2:2][2].flatten() + 0.5 * (Z.max() + Z.min())
            # Comment or uncomment following both lines to test the fake bounding box:
            for xb, yb, zb in zip(Xb, Yb, Zb):
                ax.plot([xb], [yb], [zb], 'w')

    unit = kwargs.get('unit', '')
    ax.set_xlabel('X' + unit)
    ax.set_ylabel('Y' + unit)
    ax.set_zlabel('Z' + unit)
    if kwargs.get('legend') is not None:
        ax.legend(handles, kwargs.get('legend'))
    else:
        ax.legend()
    callback = kwargs.get('callback')
    if callback is not None:
        kwargs['axes'] = ax  # TODO: make this handling consistent across the library
        callback(**kwargs)

# legacy name for plotPose3d
def plotTrajectories(dataDict, bodyIds, include, exclude, **kwargs):
    kwargs['bodyIds'] = bodyIds
    kwargs['include'] = include
    kwargs['exclude'] = exclude 
    plotPose3dMultiChannel(dataDict, **kwargs)
    


# -*- coding: utf-8 -*-
"""
Copyright (C) 2019 Event-driven Perception for Robotics
Authors: Sim Bamford
        Massimiliano Iacono

This program is free software: you can redistribute it and/or modify it under 
the terms of the GNU General Public License as published by the Free Software 
Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY 
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
PARTICULAR PURPOSE.  See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with 
this program. If not, see <https://www.gnu.org/licenses/>.

Intended as part of bimvee (Batch Import, Manipulation, Visualisation and Export of Events etc)
Functionality for serving images which represent the data at a certain time
(or given a certain time window). 
The intended use case is to support video-like playback of data
There is a generic Visualiser class - it contains a dataType dict. 
This is subclassed for each supported dataType (e.g. dvs, frame, pose etc)
Each subclass should implement basic methods: get_dims, get_frame, get_colorfmt
Implementing set_data allows the visualiser to do some preparation for visualisation
when it receives new data.

In general get_frame takes two args: time, and time_window.
In general one could think about representing data in an interpolated way or not.
For example, with poses, imu etc, one could interpolate between samples, 
or one could simply choose the sample which is nearest in time.
Likewise for frames. 
The time_window parameter says how much data to take around the sample point for event-type data.
It might be possible to develop visualisations for other types of data that make use of the concept of a time window. 

colorfmt is a choice between luminance and rgb. 
If luminance, then the frame returned should have dim 2 = 3.
Nothing stops the calling function from applying a color mask to an output in luminance format.      
"""

import numpy as np


class Visualiser:
    
    __data = None
    data_type = None

    def __init__(self, data):
        self.set_data(data)

    def set_data(self, data):
        self.__data = {}
        self.__data.update(data)
        
    def get_frame(self, time, timeWindow, **kwargs):
        return np.zeros((1, 1), dtype=np.uint8)

    def get_b_box(self, time, with_labels=True):
        return [[0, 0, 0, 0]]

    def get_colorfmt(self):
        return 'luminance'
    
    def get_settings(self):
        return {}




# -*- coding: utf-8 -*-
"""
Copyright (C) 2020 Event-driven Perception for Robotics
Authors: Sim Bamford
This program is free software: you can redistribute it and/or modify it under 
the terms of the GNU General Public License as published by the Free Software 
Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY 
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
PARTICULAR PURPOSE.  See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with 
this program. If not, see <https://www.gnu.org/licenses/>.

Intended as part of bimvee (Batch Import, Manipulation, Visualisation and Export of Events etc)

importFrames imports a folder full of (timestamped) frames saved as image files.  
It might make sense to generalise this using e.g. https://github.com/soft-matter/pims
but on day one this imports frames in the format exported by:
https://github.com/uzh-rpg/rpg_e2vid
i.e. a folder full of png, also containing a 'timestamps.txt' file, with a timestamp
in seconds written on each line. 

Returns a dict:
{'info': {'filePathOrName': str},
 'data': {
     channel0: {
         frame: {
             "ts": numpy array of float - seconds
             "frames": a list of numpy arrays where dim 0 = y (increasing downwards)

"""

#%%

import re
import numpy as np
import os
from tqdm import tqdm
import struct
import imageio

# local imports
from .timestamps import zeroTimestampsForAChannel, rezeroTimestampsForAnImportedDict

def getOrInsertDefault(inDict, arg, default):
    # get an arg from a dict.
    # If the the dict doesn't contain the arg, return the default, 
    # and also insert the default into the dict
    value = inDict.get(arg, default)
    if value == default:
        inDict[arg] = default
    return value

def importFrames(**kwargs):
    path = getOrInsertDefault(kwargs, 'filePathOrName', '.')
    print('importFrames trying path: ' + path)
    if not os.path.exists(path):
        raise FileNotFoundError("path not found.")
    if not os.path.isdir(path):
        raise FileNotFoundError("path is not a directory.")
    files = sorted(os.listdir(path))
    # TODO: trusting the os to sort the files may not work
    frames = []
    for file in tqdm(files):
        filePathAndName = os.path.join(path, file) 
        if file == 'timestamps.txt': # todo: is there a more general form?
            ts = np.loadtxt(filePathAndName)
        elif os.path.isfile(filePathAndName):
            frames.append(imageio.imread(filePathAndName))
    channelDict = {'frame': 
                       {'ts': ts,
                        'frames': frames}}
    if kwargs.get('zeroTime', kwargs.get('zeroTimestamps', True)):
        zeroTimestampsForAChannel(channelDict)
    importedDict = {
        'info': kwargs,
        'data': {'ch0': channelDict}
        }
    importedDict['info']['fileFormat'] = 'imagefolder'
    if kwargs.get('zeroTime', kwargs.get('zeroTimestamps', True)):
        rezeroTimestampsForAnImportedDict(importedDict)
    return importedDict
        
    
    


# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()
    
setup(
  name = 'bimvee',
  packages=['bimvee', 'bimvee.importRosbag', 'bimvee.importRosbag.importRosbag', 'bimvee.importRosbag.importRosbag.messageTypes', 'bimvee.visualisers'],
  version = '1.0.23',
  license='gpl',
  description = 'Batch Import, Manipulation, Visualisation and Export of Events etc',
  long_description=long_description,
  long_description_content_type='text/markdown',
  author = 'Event-driven Perception for Robotics group at Istituto Italiano di Tecnologia: Simeon Bamford, Suman Ghosh, Aiko Dinale, Massimiliano Iacono, Ander Arriandiaga, etc',
  author_email = 'simbamford@gmail.com',
  url = 'https://github.com/event-driven-robotics/bimvee',
  download_url = 'https://github.com/event-driven-robotics/bimvee_pkg/archive/v1.0.tar.gz',
  keywords = ['event', 'event camera', 'event-based', 'event-driven', 'spike', 'dvs', 'dynamic vision sensor', 'neuromorphic', 'aer', 'address-event representation' 'spiking neural network', 'davis', 'atis', 'celex' ],
  install_requires=[
          'numpy',
          'tqdm',
          'setuptools',
          'matplotlib',
          'seaborn',
          'imageio',
          'hickle',
          'opencv-python',
          'dv'
      ],
  classifiers=[
    'Development Status :: 3 - Alpha',
    'Intended Audience :: Developers',
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: GNU General Public License (GPL)',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.7',
  ],
)



# -*- coding: utf-8 -*-
"""
Copyright (C) 2019 Event-driven Perception for Robotics
Authors: Sim Bamford
        Massimiliano Iacono

This program is free software: you can redistribute it and/or modify it under 
the terms of the GNU General Public License as published by the Free Software 
Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY 
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
PARTICULAR PURPOSE.  See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with 
this program. If not, see <https://www.gnu.org/licenses/>.

Intended as part of bimvee (Batch Import, Manipulation, Visualisation and Export of Events etc)
Functionality for serving images which represent the data at a certain time
(or given a certain time window). 
The intended use case is to support video-like playback of data
There is a generic Visualiser class - it contains a dataType dict. 
This is subclassed for each supported dataType (e.g. dvs, frame, pose etc)
Each subclass should implement basic methods: get_dims, get_frame, get_colorfmt
Implementing set_data allows the visualiser to do some preparation for visualisation
when it receives new data.

VisualiserImu is for visualising imu data. A single 'cursor' is plotted, with
orthogonal unit lines pointing in x (red) y (green) and z (blue) directions.
The position of the cursor w.r.t. the centre of the visualisation indicates
linear acceleration; the orientation of the cursor indicates angular velocity.
The gain of the orientation can be altered using the kwarg 'rotation_scale' -
this interprets an input on a scale from 0 to 100 (default 50) as a multiplier
exponentially distributed in the range 0.01 through 1 (default) to 100.
The 'smoothing' kwarg (default = 0) increases the number of samples either side
which are averaged to produce the sample visualised. The input, expected to be
in the range 0 - 100, is interpretted as an exponentially distributed sample number
reaching around 117 samples each side, i.e. 235 samples in total.
"""

import numpy as np
import math
from scipy import linalg

# Local imports
from ..split import splitByLabel
from .visualiserBase import Visualiser


# A function intended to find the nearest timestamp
# adapted from https://stackoverflow.com/questions/2566412/find-nearest-value-in-numpy-array
def findNearest(array, value):
    idx = np.searchsorted(array, value)  # side="left" param is the default
    if idx > 0 and ( \
                    idx == len(array) or \
                    math.fabs(value - array[idx - 1]) < math.fabs(value - array[idx])):
        return idx - 1
    else:
        return idx


def angularVelocityToRotMat(angV):
    try:
        return linalg.expm(np.array(
            [[0, -angV[2], angV[1]],
             [angV[2], 0, -angV[0]],
             [-angV[1], angV[0], 0]]))
    except OverflowError:
        print('rotation conversion overflow')
        return np.eye(3)


# %% Two helper functions for pose visualiser

# adapted from https://stackoverflow.com/questions/50387606/python-draw-line-between-two-coordinates-in-a-matrix
def draw_line(mat, x0, y0, x1, y1):
    if (x0, y0) == (x1, y1):
        if x0 >= 0 and x0 < mat.shape[1] and y0 >= 0 and y0 < mat.shape[0]:
            mat[x0, y0] = 255
        return
    # Swap axes if Y slope is smaller than X slope
    transpose = abs(x1 - x0) < abs(y1 - y0)
    if transpose:
        mat = mat.T
        x0, y0, x1, y1 = y0, x0, y1, x1
    # Swap line direction to go left-to-right if necessary
    if x0 > x1:
        x0, y0, x1, y1 = x1, y1, x0, y0
    # Compute intermediate coordinates using line equation
    x = np.arange(x0, x1 + 1)
    y = np.round(((y1 - y0) / (x1 - x0)) * (x - x0) + y0).astype(x.dtype)
    # Write intermediate coordinates
    toKeep = np.logical_and(x >= 0,
                            np.logical_and(x < mat.shape[1],
                                           np.logical_and(y >= 0, y < mat.shape[0])))
    mat[y[toKeep], x[toKeep]] = 255


def rotateUnitVectors(rotM, unitLength=1.0):
    xVec = np.expand_dims(np.array([unitLength, 0, 0]), axis=1)
    yVec = np.expand_dims(np.array([0, unitLength, 0]), axis=1)
    zVec = np.expand_dims(np.array([0, 0, unitLength]), axis=1)
    allVecs = np.concatenate((xVec, yVec, zVec), axis=1)
    return rotM.dot(allVecs)


class VisualiserImu(Visualiser):
    renderX = 200  # TODO Hardcoded
    renderY = 200
    labels = None
    data_type = 'imu'

    def __init__(self, data):
        self.set_data(data)
        self.smallestRenderDim = min(self.renderX, self.renderY)

    '''
    Offset and scale the pose translations so that they all fit into the volume:
        x [-0.5:0.5]
        y[-0.5:0.5]
        z[1:2]
    '''

    def set_data(self, data):
        # scale and offset point data so that it remains proportional 
        # but stays in the range 0-1 for all dimensions
        minX = np.min(data['acc'][:, 0])
        maxX = np.max(data['acc'][:, 0])
        minY = np.min(data['acc'][:, 1])
        maxY = np.max(data['acc'][:, 1])
        minZ = np.min(data['acc'][:, 2])
        maxZ = np.max(data['acc'][:, 2])
        largestDim = max(maxX, -minX, maxY, -minY, maxZ, -minZ)
        if largestDim == 0:
            largestDim = 1

        pointScaled = data['acc'] / largestDim / 2 + 0.5
        self.__data = {'ts': data['ts'],
                       'point': pointScaled,
                       'rotation': data['angV']}

    def project3dTo2d(self, x=0, y=0, z=0, **kwargs):
        smallestRenderDim = kwargs.get('smallestRenderDim', 1)
        windowFill = kwargs.get('windowFill', 0.9)
        if kwargs.get('perspective', True):
            # Move z out by 1, so that the data is between 1 and 2 distant in z
            # x and y are in range 0-1, so they get shifted to be centred around 0 during this operation
            x = (x - 0.5) / (z + 1) + 0.5
            y = (y - 0.5) / (z + 1) + 0.5
        x = (x * windowFill + (1 - windowFill) / 2) * smallestRenderDim
        y = (y * windowFill + (1 - windowFill) / 2) * smallestRenderDim
        x = x.astype(int)
        y = y.astype(int)
        return x, y

    def project_pose(self, point, rotation, image, **kwargs):
        if point is None:
            return image
        # Unpack
        pointX = point[0]
        pointY = point[1]
        pointZ = point[2]
        # Project the location
        kwargs['smallestRenderDim'] = self.smallestRenderDim
        rotationScale = kwargs.get('rotation_scale', 50)
        rotationScale = 10 ** ((rotationScale - 50) / 25)
        projX, projY = self.project3dTo2d(x=pointX, y=pointY, z=pointZ, **kwargs)
        rotMat = angularVelocityToRotMat(rotation * rotationScale)
        rotatedUnitVectors = rotateUnitVectors(rotMat, unitLength=0.25)
        # aVectorCoordB means the Bth coordinate of unit vector originally directed towards a
        xVectorCoordX = pointX + rotatedUnitVectors[0, 0]
        xVectorCoordY = pointY + rotatedUnitVectors[1, 0]
        xVectorCoordZ = pointZ + rotatedUnitVectors[2, 0]
        yVectorCoordX = pointX + rotatedUnitVectors[0, 1]
        yVectorCoordY = pointY + rotatedUnitVectors[1, 1]
        yVectorCoordZ = pointZ + rotatedUnitVectors[2, 1]
        zVectorCoordX = pointX + rotatedUnitVectors[0, 2]
        zVectorCoordY = pointY + rotatedUnitVectors[1, 2]
        zVectorCoordZ = pointZ + rotatedUnitVectors[2, 2]
        xVectorProjX, xVectorProjY = self.project3dTo2d(x=xVectorCoordX, y=xVectorCoordY, z=xVectorCoordZ, **kwargs)
        yVectorProjX, yVectorProjY = self.project3dTo2d(x=yVectorCoordX, y=yVectorCoordY, z=yVectorCoordZ, **kwargs)
        zVectorProjX, zVectorProjY = self.project3dTo2d(x=zVectorCoordX, y=zVectorCoordY, z=zVectorCoordZ, **kwargs)
        draw_line(image[:, :, 0], projX, projY, xVectorProjX, xVectorProjY)
        draw_line(image[:, :, 1], projX, projY, yVectorProjX, yVectorProjY)
        draw_line(image[:, :, 2], projX, projY, zVectorProjX, zVectorProjY)
        return image

    def get_frame(self, time, timeWindow, **kwargs):
        data = self.__data
        if data is None:
            print('Warning: data is not set')
            return np.zeros((1, 1), dtype=np.uint8)  # This should not happen
        image = np.zeros((self.renderY, self.renderX, 3), dtype=np.uint8)
        # Put a grey box around the edge of the image
        image[0, :, :] = 128
        image[-1, :, :] = 128
        image[:, 0, :] = 128
        image[:, -1, :] = 128
        # Put a grey crosshair in the centre of the image
        rY = self.renderY
        rX = self.renderX
        chp = 20  # Cross Hair Proportion for following expression
        image[int(rY / 2 - rY / chp): int(rY / 2 + rY / chp), int(rX / 2), :] = 128
        image[int(rY / 2), int(rX / 2 - rX / chp): int(rX / 2 + rX / chp), :] = 128
        # Note, for the following, this follows library function pose6qInterp, but is broken out here, because of slight behavioural differences.
        idxPre = np.searchsorted(data['ts'], time, side='right') - 1
        timePre = data['ts'][idxPre]
        if idxPre < 0 or (idxPre >= len(data['ts']) - 1):
            # In this edge-case of the time at the beginning or end,
            # don't show any pose
            return image
        elif timePre == time:
            # In this edge-case of desired time == timestamp, there is no need
            # to interpolate
            poseIdx = idxPre
        else:
            poseIdx = findNearest(data['ts'], time)

        smoothing = int(kwargs.get('smoothing', 0))
        if smoothing:
            smoothing = int(1.1 ** smoothing / 2)
            firstIdx = max(0, poseIdx - smoothing)
            lastIdx = min(len(data['ts']), poseIdx + smoothing + 1)
            point = np.mean(data['point'][firstIdx:lastIdx, :], axis=0)
            rotation = np.mean(data['rotation'][firstIdx:lastIdx, :], axis=0)
        else:
            point = data['point'][poseIdx, :]
            rotation = data['rotation'][poseIdx, :]
        image = self.project_pose(point, rotation, image, **kwargs)

        return image

    def get_dims(self):
        return self.renderX, self.renderY

    def get_colorfmt(self):
        return 'rgb'

    def get_settings(self):
        settings = {'perspective': {'type': 'boolean',
                                    'default': True
                                    },
                    'rotation_scale': {'type': 'range',
                                       'default': 50,
                                       'min': 0,
                                       'max': 100,
                                       'step': 1
                                       },
                    'smoothing': {'type': 'range',
                                  'default': 0,
                                  'min': 0,
                                  'max': 100,
                                  'step': 1
                                  }}
        return settings



# -*- coding: utf-8 -*-
"""
Copyright (C) 2020 Event-driven Perception for Robotics
Author: Sim Bamford

This program is free software: you can redistribute it and/or modify it under 
the terms of the GNU General Public License as published by the Free Software 
Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY 
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
PARTICULAR PURPOSE.  See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with 
this program. If not, see <https://www.gnu.org/licenses/>.

Intended as part of bimvee (Batch Import, Manipulation, Visualisation and Export of Events etc)
Basic manipulations specific to event streams
"""

import numpy as np
from tqdm import trange
import cv2

from bimvee.plotDvsContrast import getEventImage
from bimvee.split import selectByBool

'''
removes events from pixels whose number of events is more than n * std above 
mean, where n is the 'threshold kwarg, with default value 3
'''
def removeHotPixels(inDict, **kwargs):
    # boilerplate to get down to dvs container
    if isinstance(inDict, list):
        return [removeHotPixels(inDictSingle, **kwargs)
                for inDictSingle in inDict]
    if not isinstance(inDict, dict):
        return inDict
    if 'ts' not in inDict:
        outDict = {}
        for key in inDict.keys():
            outDict[key] = removeHotPixels(inDict[key], **kwargs)
        return outDict
    # From this point onwards, it's a data-type container
    if 'pol' not in inDict:
        return
    # From this point onwards, it's a dvs container
    events = inDict
    eventImage = getEventImage(events, contrast=np.inf, polarised=False)
    contrast1d = eventImage.flatten()
    mean = np.mean(contrast1d)
    std = np.std(contrast1d)
    threshold = mean + kwargs.get('threshold', 3) * std
    (y, x) = np.where(eventImage > threshold)
    dimY = kwargs.get('dimY', events.get('dimY', events['y'].max() + 1))
    #dimX = kwargs.get('dimX', events.get('dimX', events['x'].max()))
    addrsToRemove = x * dimY + y
    eventAddrs = events['x'] * dimY + events['y']
    toKeep = np.logical_not(np.isin(eventAddrs, addrsToRemove))
    return selectByBool(events, toKeep)

''' 
Iterates event by event, using a prevTs array to keep track of each pixel. 
From a brief trial, dissecting into different arrays, 
doing the filter and then merging again is quite a lot slower than this. 
'''
def refractoryPeriod(inDict, refractoryPeriod=0.001, **kwargs):
    # boilerplate to get down to dvs container
    if isinstance(inDict, list):
        return [refractoryPeriod(inDictSingle, **kwargs)
                for inDictSingle in inDict]
    if not isinstance(inDict, dict):
        return inDict
    if 'ts' not in inDict:
        outDict = {}
        for key in inDict.keys():
            outDict[key] = refractoryPeriod(inDict[key], **kwargs)
        return outDict
    # From this point onwards, it's a data-type container
    if 'pol' not in inDict:
        return
    # From this point onwards, it's a dvs container
    events = inDict
    ts = events['ts']
    x = events['y']
    y = events['x']
    numEvents = len(ts)
    maxX = x.max()
    maxY = y.max()
    prevTs = np.zeros((maxY+1, maxX+1))
    toKeep = np.ones((numEvents), dtype=bool)
    for idx in trange(numEvents, leave=True, position=0):
        if ts[idx] >= prevTs[y[idx], x[idx]] + refractoryPeriod: 
            prevTs[y[idx], x[idx]] = ts[idx]
        else:
            toKeep[idx] = False
    outDict = selectByBool(inDict, toKeep)
    return outDict

'''
if dict does not already have dimX and dimY fields, then add them, using the
maximum value in each dimension.
NOTE: This is a hap-hazard method which relies on pixels actually producing 
events; only to be used if camera dimension is not otherwise available.
'''
def findDims(inDict):
    if 'dimX' in inDict and 'dimY' in inDict:
        return inDict
    outDict = inDict.copy()
    if 'dimX' not in inDict:
        outDict['dimX'] = np.max(inDict['x']) + 1
    if 'dimY' not in inDict:
        outDict['dimY'] = np.max(inDict['y']) + 1
    return outDict

'''
Converts the underlying representation of address-events to a single ndarray
n x 3, where the first col is x, the second col is y, and the third col is all 
ones; the dtype is float. The field is called 'xyh'.
x and y fields become 1d views of the appropriate rows.
'''
def convertToHomogeneousCoords(inDict):
    outDict = inDict.copy()
    outDict['xyh'] = np.concatenate((
        inDict['x'][:, np.newaxis].astype(float),
        inDict['y'][:, np.newaxis].astype(float),
        np.ones((inDict['x'].shape[0], 1), dtype=float)
        ), axis=1)
    outDict['x'] = outDict['xyh'][:, 0]
    outDict['y'] = outDict['xyh'][:, 1]
    return outDict

'''
Uses opencv functions to create an undistortion map and undistort events.
k is the intrinsic matrix; d is the distortion coefficients.
Returns a new container with the events undistorted.
By default, events get turned into float64; keep them as int16 with kwarg 'asInt'=True
By default, it uses the same intrinsic matrix in the remapping - change this
by passing in the kwarg 'kNew'
'''
def undistortEvents(inDict, k, d, **kwargs):
    inDict = findDims(inDict)
    kNew = kwargs.get('kNew', k)
    yGrid, xGrid = np.meshgrid(range(inDict['dimY']), range(inDict['dimX']))
    xGrid = xGrid.reshape(-1).astype(float)
    yGrid = yGrid.reshape(-1).astype(float)
    xyGrid = np.concatenate((xGrid[:, np.newaxis], yGrid[:, np.newaxis]), axis=1)
    undistortedPoints = cv2.undistortPoints(xyGrid, k, d, None, kNew)
    undistortionMap = undistortedPoints.reshape(inDict['dimX'], inDict['dimY'], 2)
    undistortionMap = np.swapaxes(undistortionMap, 0, 1)
    xy = undistortionMap[inDict['y'], inDict['x'], :]
    if kwargs.get('asInt', False):
        xy = np.round(xy).astype('int')
    outDict = inDict.copy()
    outDict['x'] = xy[:, 0]
    outDict['y'] = xy[:, 1]
    return outDict


# -*- coding: utf-8 -*-
"""
Copyright (C) 2020 Event-driven Perception for Robotics
Author: Sim Bamford

This program is free software: you can redistribute it and/or modify it under 
the terms of the GNU General Public License as published by the Free Software 
Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY 
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
PARTICULAR PURPOSE.  See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with 
this program. If not, see <https://www.gnu.org/licenses/>.

Intended as part of bimvee (Batch Import, Manipulation, Visualisation and Export of Events etc)

A Container class which encapsulates some basic container-level manipulations.

Firstly, the loosely enforced 3-tier hierarchy of data-channel-datatype within 
an imported dict has the advantage that channels can help organise data but 
it has the disadvantage that when there is just a single dict for each datatype, 
the user may not know or care about the channel names, yet needs them in order
to navigate the data. 
Therefore the getDataType() method allows the user to get the dict for a chosen
dataType providing only that there's a single example. 
"""

class Container():
    
    def __init__(self, container):
       self.container = container
     
    # This code assumes the 3-level hierarchy    
    def getDataType(self, dataType):
        dataTypeDicts = []
        for channelKey in self.container['data'].keys():
            for dataTypeKey in self.container['data'][channelKey]:
                if dataTypeKey == dataType:
                    dataTypeDicts.append(self.container['data'][channelKey][dataTypeKey])
        if len(dataTypeDicts) == 0:
            raise ValueError('No dicts found for dataType ' + dataType)
        if len(dataTypeDicts) > 1:
            raise ValueError('More than one dict found for dataType: ' + dataType)        
        return dataTypeDicts[0]

    def getAllDataOfType(self, dataType):
        dataTypeDicts = {}
        for channelKey in self.container['data'].keys():
            for dataTypeKey in self.container['data'][channelKey]:
                if dataTypeKey == dataType:
                    dataTypeDicts[channelKey] = self.container['data'][channelKey][dataTypeKey]
        return dataTypeDicts




# -*- coding: utf-8 -*-

"""
Copyright (C) 2019 Event-driven Perception for Robotics
Authors: Sim Bamford
This program is free software: you can redistribute it and/or modify it under 
the terms of the GNU General Public License as published by the Free Software 
Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY 
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
PARTICULAR PURPOSE.  See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with 
this program. If not, see <https://www.gnu.org/licenses/>.

Intended as part of importRosbag.

The importRosbag function receives a list of messages and returns
a dict with one field for each data field in the message, where the field
will contain an appropriate iterable to contain the interpretted contents of each message.

This function imports the ros message type defined at:

https://github.com/uzh-rpg/rpg_esim/blob/master/event_camera_simulator/esim_msgs/msg/OpticFlow.msg

"""

#%%

from tqdm import tqdm
import numpy as np

from .common import unpackRosFloat32Array, unpackRosUint32, \
                    unpackRosTimestamp, unpackRosString

def importTopic(msgs, **kwargs):
    disable_bar = kwargs.get('disable_bar')
    tsAll = []
    flowMaps = []
    for msg in tqdm(msgs, disable=disable_bar):
        
        data = msg['data']
        ptr = 0
        seq, ptr = unpackRosUint32(data, ptr) # Not used
        ts, ptr = unpackRosTimestamp(data, ptr)
        frame_id, ptr = unpackRosString(data, ptr) # Not used
        height, ptr = unpackRosUint32(data, ptr)
        width, ptr = unpackRosUint32(data, ptr) 

        if width > 0 and height > 0:
            arraySize, ptr = unpackRosUint32(data, ptr)
            #assert arraySize == width*height
            flowMapX, ptr = unpackRosFloat32Array(data, width*height, ptr)
            arraySize, ptr = unpackRosUint32(data, ptr)
            #assert arraySize == width*height
            flowMapY, ptr = unpackRosFloat32Array(data, width*height, ptr)
            flowMap = np.concatenate((flowMapX.reshape(height, width, 1), 
                                      flowMapY.reshape(height, width, 1)), 
                                     axis=2)
            flowMaps.append(flowMap)
            tsAll.append(ts)
    if not tsAll:
        return None
    outDict = {
        'ts': np.array(tsAll, dtype=np.float64),
        'flowMaps': flowMaps,
        }
    return outDict


